# -*- coding: utf-8 -*-
"""This module defines functions for parsing and writing `PDB files`_.

.. _PDB files: http://www.wwpdb.org/documentation/format32/v3.2.html"""

from collections import defaultdict
import os.path
import time
from numbers import Integral

import numpy as np

from prody.atomic import AtomGroup, Atom, Selection
from prody.atomic import flags, AAMAP
from prody.atomic import ATOMIC_FIELDS
from prody.utilities import openFile, isListLike
from prody.utilities.misctools import decToHybrid36, hybrid36ToDec, packmolRenumChains
from prody import LOGGER, SETTINGS

from .header import getHeaderDict, buildBiomolecules, assignSecstr, isHelix, isSheet
from .localpdb import fetchPDB
from .ciffile import parseMMCIF
from .emdfile import parseEMD

__all__ = ['parsePDBStream', 'parsePDB', 'parseChainsList', 'parsePQR',
           'writePDBStream', 'writePDB', 'writeChainsList', 'writePQR',
           'writePQRStream']

MAX_N_ATOM = 99999 
MAX_N_RES = 9999

long_id_check_str = "len(pdb) == %d and pdb.startswith('pdb_') and pdb[3] == '_'"

class PDBParseError(Exception):
    pass

_parsePQRdoc = """
    :arg title: title of the :class:`.AtomGroup` instance, default is the
        PDB filename or PDB identifier
    :type title: str

    :arg ag: :class:`.AtomGroup` instance for storing data parsed from PDB
        file, number of atoms in *ag* and number of atoms parsed from the PDB
        file must be the same and atoms in *ag* and those in PDB file must
        be in the same order.  Non-coordinate data stored in *ag* will be
        overwritten with those parsed from the file.
    :type ag: :class:`.AtomGroup`

    :arg chain: chain identifiers for parsing specific chains, e.g.
        ``chain='A'``, ``chain='B'``, ``chain='DE'``, by default all
        chains are parsed
    :type chain: str

    :arg subset: a predefined keyword to parse subset of atoms, valid keywords
        are ``'calpha'`` (``'ca'``), ``'backbone'`` (``'bb'``), or **None**
        (read all atoms), e.g. ``subset='bb'``
    :type subset: str
"""

_parsePDBdoc = _parsePQRdoc + """
    :arg model: model index or None (read all models), e.g. ``model=10``
    :type model: int, list

    :arg header: if **True** PDB header content will be parsed and returned
    :type header: bool

    :arg altloc: if a location indicator is passed, such as ``'A'`` or ``'B'``,
         only indicated alternate locations will be parsed as the single
         coordinate set of the AtomGroup,  if *altloc* is set ``'all'`` then all
         alternate locations will be parsed and each will be appended as a
         distinct coordinate set, default is ``"A"``
    :type altloc: str

    :arg biomol: if **True**, biomolecules are obtained by transforming the
        coordinates using information from header section will be returned.
        This option uses :func:`.buildBiomolecules` and as noted there, 
        atoms in biomolecules are ordered according to the original chain 
        IDs. Chains may have the same chain ID, in which case they are given 
        different segment names.
        Default is **False**
    :type biomol: bool

    :arg secondary: if **True**, secondary structure information from header
        section will be assigned to atoms.
        Default is **False**
    :type secondary: bool

    If ``model=0`` and ``header=True``, return header dictionary only.

    """

_PDBSubsets = {'ca': 'ca', 'calpha': 'ca', 'bb': 'bb', 'backbone': 'bb'}

def parsePDB(*pdb, **kwargs):
    """Returns an :class:`.AtomGroup` and/or dictionary containing header data
    parsed from a PDB file.

    This function extends :func:`.parsePDBStream`.

    See :ref:`parsepdb` for a detailed usage example.

    :arg pdb: one PDB identifier or filename, or a list of them.
        If needed, PDB files are downloaded using :func:`.fetchPDB()` function.
    
    You can also provide arguments that you would like passed on to fetchPDB().

    :arg extend_biomol: whether to extend the list of results with a list
        rather than appending, which can create a mixed list, 
        especially when biomol=True.
        Default value is False to reproduce previous behaviour.
        This value is ignored when result is not a list (header=True or model=0).
    :type extend_biomol: bool 

    Please note that resnames are only taken as 3 characters and chids can be 2
    by default. Hence, TIP3S is split into resname TIP and chid 3S.

    This behaviour can be changed with the following two arguments:

    :arg long_resname: whether to parse longer resnames instead of cutting at 3 chars
    :type long_resname: bool

    :arg long_chid: whether to parse longer chain ids instead of cutting at 2 chars
    :type long_chid: bool
    """
    extend_biomol = kwargs.pop('extend_biomol', False)

    n_pdb = len(pdb)
    if n_pdb == 0:
        raise ValueError('Please provide a PDB ID or filename')

    if n_pdb == 1 and isListLike(pdb[0]) or isinstance(pdb[0], dict):
        pdb = pdb[0]
        n_pdb = len(pdb)

    if n_pdb == 1:
        return _parsePDB(pdb[0], **kwargs)
    else:
        results = []
        lstkwargs = {}
        for key in kwargs:
            argval = kwargs.get(key)
            if np.isscalar(argval):
                argval = [argval]*n_pdb
            lstkwargs[key] = argval

        start = time.time()
        LOGGER.progress('Retrieving {0} PDB structures...'
                    .format(n_pdb), n_pdb, '_prody_parsePDB')
        for i, p in enumerate(pdb):
            kwargs = {}
            for key in lstkwargs:
                kwargs[key] = lstkwargs[key][i]
            c = kwargs.get('chain','')
            LOGGER.update(i, 'Retrieving {0}...'.format(p+c), 
                          label='_prody_parsePDB')
            result = _parsePDB(p, **kwargs)
            if not isinstance(result, tuple):
                if isinstance(result, dict):
                    result = (None, result)
                else:
                    result = (result, None)
            results.append(result)

        results = list(zip(*results))
        LOGGER.finish()

        for i in reversed(range(len(results))):
            if all(j is None for j in results[i]):
                results.pop(i)
        if len(results) == 1:
            results = results[0]
        results = list(results)

        model = kwargs.get('model')
        header = kwargs.get('header', False)
        if model != 0 and header:
            numPdbs = len(results[0])
        else:
            numPdbs = len(results)

            if extend_biomol:
                results_old = results
                results = []
                for entry in results_old:
                    if isinstance(entry, AtomGroup):
                        results.append(entry)
                    else:
                        results.extend(entry)

        LOGGER.info('{0} PDBs were parsed in {1:.2f}s.'
                     .format(numPdbs, time.time()-start))

        return results

def _getPDBid(pdb):
    l = len(pdb)
    if l == 4 or eval(long_id_check_str % 12):
        pdbid, chain = pdb, ''
    elif l == 5 or eval(long_id_check_str % 13):
        pdbid = pdb[:-1]; chain = pdb[-1]
    elif ':' in pdb:
        i = pdb.find(':')
        pdbid = pdb[:i]; chain = pdb[i+1:]
    else:
        raise IOError('{0} is not a valid filename or a valid PDB '
                      'identifier.'.format(pdb))
    if not (pdbid.isalnum() or len(pdbid) == 12 and pdbid.startswith('pdb_') and pdbid[3] == '_'):
        raise IOError('{0} is not a valid filename or a valid PDB '
                      'identifier.'.format(pdb))
    if chain != '' and not chain.isalnum():
        raise IOError('{0} is not a valid chain identifier.'.format(chain))
    return pdbid, chain

def _parsePDB(pdb, **kwargs):
    title = kwargs.get('title', None)
    chain = ''
    if not os.path.isfile(pdb):
        pdb, chain = _getPDBid(pdb)
        if title is None:
            title = pdb
            kwargs['title'] = title
        filename = fetchPDB(pdb, **kwargs)
        if filename is None:
            try:
                if not (eval(long_id_check_str % 12) or eval(long_id_check_str % 13)):
                    LOGGER.warn("Trying to parse mmCIF file instead")
                chain = kwargs.pop('chain', chain)
                return parseMMCIF(pdb+chain, **kwargs)
            except OSError:
                try:
                    LOGGER.warn("Trying to parse EMD file instead")
                    chain = kwargs.pop('chain', chain)
                    return parseEMD(pdb+chain, **kwargs)
                except:
                    raise IOError('PDB file for {0} could not be parsed.'
                                .format(pdb))
        pdb = filename
    if title is None:
        title, ext = os.path.splitext(os.path.split(pdb)[1])
        if ext == '.gz':
            title, ext = os.path.splitext(title)
        if len(title) == 7 and title.startswith('pdb'):
            title = title[3:]
        kwargs['title'] = title

    if pdb.endswith('.pdb') or pdb.endswith('.pdb.gz'):
        stream = openFile(pdb, 'rt')
        if chain != '':
            kwargs['chain'] = chain
        result = parsePDBStream(stream, **kwargs)
        stream.close()
        return result
    else:
        try:
            LOGGER.warn("Trying to parse as mmCIF file instead")
            return parseMMCIF(pdb, **kwargs)
        except KeyError:
            try:
                LOGGER.warn("Trying to parse as EMD file instead")
                return parseEMD(pdb, **kwargs)
            except ValueError:
                LOGGER.warn("Could not parse anything so returning None")
                return None
            except:
                raise IOError('PDB file for {0} could not be downloaded.'
                              .format(pdb))

parsePDB.__doc__ += _parsePDBdoc

def parsePDBStream(stream, **kwargs):
    """Returns an :class:`.AtomGroup` and/or dictionary containing header data
    parsed from a stream of PDB lines.

    :arg stream: Anything that implements the method ``readlines``
        (e.g. :class:`file`, buffer, stdin)""" 

    model = kwargs.get('model')
    header = kwargs.get('header', False)
    assert isinstance(header, bool), 'header must be a boolean'
    chain = kwargs.get('chain')
    subset = kwargs.get('subset')
    altloc = kwargs.get('altloc', 'A')
    packmol = kwargs.get('packmol', False)

    auto_bonds = SETTINGS.get('auto_bonds')
    get_bonds = kwargs.get('bonds', auto_bonds)

    long_resname = kwargs.get('long_resname')
    long_chid = kwargs.get('long_chid')
    strip_icodes = kwargs.get('strip_icodes', True)

    if model is not None:
        if isinstance(model, Integral):
            if model < 0:
                raise ValueError('model must be greater than 0')
        else:
            raise TypeError('model must be an integer, {0} is invalid'
                            .format(str(model)))
    title_suffix = ''
    if subset:
        try:
            subset = _PDBSubsets[subset.lower()]
        except AttributeError:
            raise TypeError('subset must be a string')
        except KeyError:
            raise ValueError('{0} is not a valid subset'
                             .format(repr(subset)))
        title_suffix = '_' + subset
    if chain is not None:
        if not isinstance(chain, str):
            raise TypeError('chain must be a string')
        elif len(chain) == 0:
            raise ValueError('chain must not be an empty string')
        title_suffix = chain + title_suffix
    ag = kwargs.pop('ag', None)
    if ag is not None:
        if not isinstance(ag, AtomGroup):
            raise TypeError('ag must be an AtomGroup instance')
        n_csets = ag.numCoordsets()
    elif model != 0:
        ag = AtomGroup(str(kwargs.get('title', 'Unknown')) + title_suffix)
        n_csets = 0

    biomol = kwargs.get('biomol', False)
    auto_secondary = SETTINGS.get('auto_secondary')
    secondary = kwargs.get('secondary', auto_secondary)
    split = 0
    hd = None
    if model != 0:
        LOGGER.timeit()
        try:
            lines = stream.readlines()
        except AttributeError as err:
            try:
                lines = stream.read().split('\n')
            except AttributeError:
                raise err
        if not len(lines):
            raise ValueError('empty PDB file or stream')
        if header or biomol or secondary:
            hd, split = getHeaderDict(lines)
        bonds = [] if get_bonds else None
        _parsePDBLines(ag, lines, split, model, chain, subset, altloc, bonds=bonds, 
                       long_resname=long_resname, long_chid=long_chid, strip_icodes=strip_icodes)
        if bonds:
            try:
                ag.setBonds(bonds)
            except ValueError:
                LOGGER.warn('Bonds read from CONECT records do not apply to subset so were not added')
        if ag.numAtoms() > 0:
            LOGGER.report('{0} atoms and {1} coordinate set(s) were '
                          'parsed in %.2fs.'.format(ag.numAtoms(),
                          ag.numCoordsets() - n_csets))
        else:
            ag = None
            LOGGER.warn('Atomic data could not be parsed, please '
                        'check the input file.')
    elif header:
        hd, split = getHeaderDict(stream)

    if ag is not None and isinstance(hd, dict):
        if secondary:
            if auto_secondary:
                try:
                    ag = assignSecstr(hd, ag)
                except ValueError:
                    pass
            else:
                ag = assignSecstr(hd, ag)
        if biomol:
            ag = buildBiomolecules(hd, ag)

            if isinstance(ag, list):
                LOGGER.info('Biomolecular transformations were applied, {0} '
                            'biomolecule(s) are returned.'.format(len(ag)))
            else:
                LOGGER.info('Biomolecular transformations were applied to the '
                            'coordinate data.')

    if packmol:
        ag = packmolRenumChains(ag)

    if model != 0:
        if header:
            return ag, hd
        else:
            return ag
    else:
        return hd

parsePDBStream.__doc__ += _parsePDBdoc


def parsePQR(filename, **kwargs):
    """Returns an :class:`.AtomGroup` containing data parsed from PDB lines.

    :arg filename: a PQR filename
    :type filename: str"""

    title = kwargs.get('title', kwargs.get('name'))
    chain = kwargs.get('chain')
    subset = kwargs.get('subset')
    long_resname = kwargs.get('long_resname')
    long_chid = kwargs.get('long_chid')
    strip_icodes = kwargs.get('strip_icodes', True)
    if not os.path.isfile(filename):
        raise IOError('No such file: {0}'.format(repr(filename)))
    if title is None:
        fn, ext = os.path.splitext(os.path.split(filename)[1])
        if ext == '.gz':
            fn, ext = os.path.splitext(fn)
        title = fn.lower()
    title_suffix = ''
    if subset:
        try:
            subset = _PDBSubsets[subset.lower()]
        except AttributeError:
            raise TypeError('subset must be a string')
        except KeyError:
            raise ValueError('{0} is not a valid subset'
                             .format(repr(subset)))
        title_suffix = '_' + subset
    if chain is not None:
        if not isinstance(chain, str):
            raise TypeError('chain must be a string')
        elif len(chain) == 0:
            raise ValueError('chain must not be an empty string')
        title_suffix = '_' + chain + title_suffix
    if 'ag' in kwargs:
        ag = kwargs['ag']
        if not isinstance(ag, AtomGroup):
            raise TypeError('ag must be an AtomGroup instance')
        n_csets = ag.numCoordsets()
    else:
        ag = AtomGroup(title + title_suffix)
        n_csets = 0

    pqr = openFile(filename, 'rt')
    lines = pqr.readlines()
    pqr.close()
    LOGGER.timeit()
    ag = _parsePDBLines(ag, lines, split=0, model=1, chain=chain,
                        subset=subset, altloc_torf=False, format='pqr', 
                        long_resname=long_resname, long_chid=long_chid,
                        strip_icodes=strip_icodes)
    if ag.numAtoms() > 0:
        LOGGER.report('{0} atoms and {1} coordinate sets were '
                      'parsed in %.2fs.'.format(ag.numAtoms(),
                      ag.numCoordsets() - n_csets))
        return ag
    else:
        return None

parsePQR.__doc__ += _parsePQRdoc

def _parsePDBLines(atomgroup, lines, split, model, chain, subset,
                   altloc_torf, format='PDB', bonds=None, 
                   long_resname=False, long_chid=False, strip_icodes=True):
    """Returns an AtomGroup. See also :func:`.parsePDBStream()`.

    :arg lines: PDB/PQR lines
    :arg split: starting index for coordinate data lines"""

    format = format.upper()
    isPDB = format == 'PDB'

    num_ters = 0

    if subset:
        if subset == 'ca':
            subset = set(('CA',))
        elif subset in 'bb':
            subset = flags.BACKBONE
        only_subset = True
        protein_resnames = flags.AMINOACIDS
    else:
        only_subset = False
    if chain is None:
        only_chains = False
    else:
        only_chains = True
    onlycoords = False
    n_atoms = atomgroup.numAtoms()
    if n_atoms > 0:
        asize = n_atoms
    else:
        asize = len(lines) - split
    addcoords = False
    if atomgroup.numCoordsets() > 0:
        addcoords = True
    alength = asize
    coordinates = np.zeros((asize, 3), dtype=float)
    atomnames = np.zeros(asize, dtype=ATOMIC_FIELDS['name'].dtype)
    resnames = np.zeros(asize, dtype=ATOMIC_FIELDS['resname'].dtype)
    resnums = np.zeros(asize, dtype=ATOMIC_FIELDS['resnum'].dtype)
    chainids = np.zeros(asize, dtype=ATOMIC_FIELDS['chain'].dtype)
    hetero = np.zeros(asize, dtype=bool)
    termini = np.zeros(asize, dtype=bool)
    altlocs = np.zeros(asize, dtype=ATOMIC_FIELDS['altloc'].dtype)
    icodes = np.zeros(asize, dtype=ATOMIC_FIELDS['icode'].dtype)
    serials = np.zeros(asize, dtype=ATOMIC_FIELDS['serial'].dtype)
    charges = np.zeros(asize, dtype=ATOMIC_FIELDS['charge'].dtype)
    if isPDB:
        segnames = np.zeros(asize, dtype=ATOMIC_FIELDS['segment'].dtype)
        elements = np.zeros(asize, dtype=ATOMIC_FIELDS['element'].dtype)
        bfactors = np.zeros(asize, dtype=ATOMIC_FIELDS['beta'].dtype)
        occupancies = np.zeros(asize, dtype=ATOMIC_FIELDS['occupancy'].dtype)
        siguij = None
    else:
        radii = np.zeros(asize, dtype=ATOMIC_FIELDS['radius'].dtype)
    anisou = None
    asize = 2000 # increase array length by this much when needed

    start = split
    stop = len(lines)
    nmodel = 0
    bonds_serial = []
    # if a specific model is requested, skip lines until that one
    if isPDB and model is not None and model != 1:
        for i in range(split, len(lines)):
            if lines[i][:5] == 'MODEL':
                nmodel += 1
                if model == nmodel:
                    start = i+1
                    stop = len(lines)
                    break
        if nmodel != model:
            raise PDBParseError('model {0} is not found'.format(model))

    if isinstance(altloc_torf, str):
        if altloc_torf == 'all':
            which_altlocs = 'all'
        elif altloc_torf.strip() != 'A':
            LOGGER.info('Parsing alternate locations {0}.'
                        .format(altloc_torf))
            which_altlocs = ' ' + ''.join(altloc_torf.split())
        else:
            which_altlocs = ' A'
        altloc_torf = False
    else:
        which_altlocs = ' A'
        altloc_torf = True

    acount = 0
    coordsets = None
    altloc = defaultdict(list)
    i = start
    END = False
    warned_5_digit = False
    warned_long_chid = False
    dec = True
    while i < stop:
        line = lines[i]
        startswith = line[0:6].strip()

        if startswith == 'ATOM' or startswith == 'HETATM':
            if isPDB:
                atomname = line[12:16].strip()

                if long_resname:
                    resname = line[17:21].strip()
                else:
                    resname = line[17:20].strip()
            else:
                fields = line.split()
                if fields[5].find('.') != -1:
                    # coords too early as no chid
                    fields.insert(4, '')
                if len(fields) != 11:
                    try:
                        fields = fields[:6] + [line[30:38].strip(), line[38:46].strip(), line[46:54].strip()] + line[54:].split()
                    except:
                        LOGGER.warn('wrong number of fields for PQR format at line %d'%i)
                        i += 1
                        continue
                atomname= fields[2]
                resname = fields[3]

            if only_subset:
                if not (atomname in subset and resname in protein_resnames):
                    i += 1
                    continue

            if isPDB:
                if long_chid:
                    chid = line[20:22].strip()
                    if len(chid) > 1 and not warned_long_chid:
                        LOGGER.warn('Parsed 2-character chid {0} continuous with resnum {1} from {2}. Please check if this was intended.'.format(
                            chid, line[17:20], line[17:22]
                        ))
                        warned_long_chid = True
                else:
                    chid = line[21].strip()
            else:
                chid = fields[4]

            if only_chains:
                if not chid in chain:
                    i += 1
                    continue

            if isPDB:
                alt = line[16]
                if alt not in which_altlocs and which_altlocs != 'all':
                    altloc[alt].append((line, i))
                    i += 1
                    continue
            else:
                alt = ' '
            try:
                if isPDB:
                    coordinates[acount, 0] = line[30:38]
                    coordinates[acount, 1] = line[38:46]
                    coordinates[acount, 2] = line[46:54]
                else:
                    coordinates[acount, 0] = fields[6]
                    coordinates[acount, 1] = fields[7]
                    coordinates[acount, 2] = fields[8]
            except:
                if acount >= n_atoms > 0:
                    if nmodel == 0:
                        raise ValueError(format + 'file and AtomGroup ag must '
                                         'have same number of atoms')
                    LOGGER.warn('Discarding model {0}, which contains {1} more '
                                'atoms than first model does.'
                                .format(nmodel+1,acount-n_atoms+1))
                    acount = 0
                    nmodel += 1
                    coordinates = np.zeros((n_atoms, 3), dtype=float)
                    if isPDB:
                        while lines[i][:6] != 'ENDMDL':
                            i += 1
                else:
                    raise PDBParseError('invalid or missing coordinate(s) at '
                                         'line {0}'.format(i+1))
            if onlycoords:
                acount += 1
                i += 1
                continue

            serial_str = line[6:11] if isPDB else fields[1]
            try:
                serials[acount] = int(serial_str)
            except ValueError:
                try:
                    isnumeric = np.all([x.isdigit() for x in serial_str])
                    if not isnumeric and serial_str == serial_str.upper():
                        serials[acount] = hybrid36ToDec(serial_str)
                    else:
                        # lower case is found in hexadecimal PDB files
                        serials[acount] = int(serial_str, 16)
                except ValueError:
                    if acount > 0:
                        LOGGER.warn('failed to parse serial number in line {0}. Assigning it by incrementing.'
                                    .format(i))
                        serials[acount] = serials[acount-1]+1
                    else:
                        LOGGER.warn('failed to parse serial number in line {0}. Assigning it as 1.'
                                    .format(i))
                        serials[acount] = 1

            altlocs[acount] = alt
            atomnames[acount] = atomname
            resnames[acount] = resname
            chainids[acount] = chid
            if isPDB:
                resnum_str = line[22:26]
                icode = line[26]
                if dec:
                    try:
                        resnum = int(resnum_str)
                    except ValueError:
                        dec = False

                    if icode.isdigit():
                        if not warned_5_digit:
                            LOGGER.warn('parsed 5 digit residue number including numeric insertion code')
                            warned_5_digit = True
                        resnum = int(str(resnum) + icode)
                    else:
                        icodes[acount] = icode

                if dec and acount > 2 and resnums[acount-2] > resnum and resnums[acount-2] >= MAX_N_RES:
                    dec = False

                if not dec:
                    resnum = resnum_str
                    try:
                        isnumeric = np.all([x.isdigit() or x==' ' for x in resnum_str])
                        if not isnumeric and resnum_str == resnum_str.upper():
                            resnum = hybrid36ToDec(resnum_str, resnum=True)
                        else:
                            # lower case is found in hexadecimal PDB files
                            resnum = int(resnum_str, 16)

                    except ValueError:
                        if acount > 0:
                            LOGGER.warn('failed to parse residue number in line {0}. Assigning it by incrementing.'
                                        .format(i))
                            resnum = resnums[acount-1]+1
                        else:
                            LOGGER.warn('failed to parse residue number in line {0}. Assigning it as 1.'
                                        .format(i))
                            resnum = 1

                resnums[acount] = resnum
            else:
                resnum = fields[5]
                if resnum[-1].isalpha():
                    icode = resnum[-1]
                else:
                    icode = ' '
                try:
                    resnums[acount] = int(resnum)
                except ValueError:
                    try:
                        resnums[acount] = int(resnum, 16)
                    except ValueError:
                        LOGGER.warn('failed to parse residue number in line {0}. Assigning it by incrementing.'
                                    .format(i))
                        resnums[acount] = resnums[acount-1]+1
                icodes[acount] = icode

            if isPDB:
                try:
                    occupancies[acount] = line[54:60]
                except:
                    LOGGER.warn('failed to parse occupancy at line {0}'
                                .format(i))
                try:
                    bfactors[acount] = line[60:66]
                except:
                    LOGGER.warn('failed to parse beta-factor at line {0}'
                                .format(i))
                hetero[acount] = startswith[0] == 'H'
                segnames[acount] = line[72:76]
                elements[acount] = line[76:78]
                try:
                    charges[acount] = int(line[79] + line[78])
                except:
                    charges[acount] = 0
            else:
                try:
                    charges[acount] = fields[9]
                except:
                    LOGGER.warn('failed to parse charge at line {0}'
                                .format(i))
                try:
                    radii[acount] = fields[10]
                except:
                    LOGGER.warn('failed to parse radius at line {0}'
                                .format(i))
            acount += 1
            if n_atoms == 0 and acount >= alength:
                # if arrays are short extend them with zeros
                alength += asize
                coordinates = np.concatenate(
                    (coordinates, np.zeros((asize, 3), float)))
                atomnames = np.concatenate((atomnames,
                    np.zeros(asize, ATOMIC_FIELDS['name'].dtype)))
                resnames = np.concatenate((resnames,
                    np.zeros(asize, ATOMIC_FIELDS['resname'].dtype)))
                resnums = np.concatenate((resnums,
                    np.zeros(asize, ATOMIC_FIELDS['resnum'].dtype)))
                chainids = np.concatenate((chainids,
                    np.zeros(asize, ATOMIC_FIELDS['chain'].dtype)))
                hetero = np.concatenate((hetero, np.zeros(asize, bool)))
                termini = np.concatenate((termini, np.zeros(asize, bool)))
                altlocs = np.concatenate((altlocs,
                    np.zeros(asize, ATOMIC_FIELDS['altloc'].dtype)))
                icodes = np.concatenate((icodes,
                    np.zeros(asize, ATOMIC_FIELDS['icode'].dtype)))
                serials = np.concatenate((serials,
                    np.zeros(asize, ATOMIC_FIELDS['serial'].dtype)))
                if isPDB:
                    bfactors = np.concatenate((bfactors,
                        np.zeros(asize, ATOMIC_FIELDS['beta'].dtype)))
                    occupancies = np.concatenate((occupancies,
                        np.zeros(asize, ATOMIC_FIELDS['occupancy'].dtype)))
                    segnames = np.concatenate((segnames,
                        np.zeros(asize, ATOMIC_FIELDS['segment'].dtype)))
                    elements = np.concatenate((elements,
                        np.zeros(asize, ATOMIC_FIELDS['element'].dtype)))
                    if anisou is not None:
                        anisou = np.concatenate((anisou, np.zeros((asize, 6), float)))
                    if siguij is not None:
                        siguij = np.concatenate((siguij, np.zeros((asize, 6),
                            ATOMIC_FIELDS['siguij'].dtype)))
                else:
                    charges = np.concatenate((charges,
                        np.zeros(asize, ATOMIC_FIELDS['charge'].dtype)))
                    radii = np.concatenate((radii,
                        np.zeros(asize, ATOMIC_FIELDS['radius'].dtype)))
        elif startswith == 'CONECT':
            if bonds is not None:
                atom_serial = line[6:11]
                bonded1_serial = line[11:16]
                bonds_serial.append([int(atom_serial), int(bonded1_serial)])

                bonded2_serial = line[16:21]
                if len(bonded2_serial.strip()):
                    bonds_serial.append([int(atom_serial), int(bonded2_serial)])

                bonded3_serial = line[21:26]
                if len(bonded3_serial.strip()):
                    bonds_serial.append([int(atom_serial), int(bonded3_serial)])

                bonded4_serial = line[26:31]  # fixed typo
                if len(bonded4_serial.strip()):
                    bonds_serial.append([int(atom_serial), int(bonded4_serial)])

        elif not onlycoords and (startswith == 'TER   ' or
            startswith.strip() == 'TER'):
            termini[acount - 1] = True
            num_ters += 1
        elif startswith == 'ENDMDL' or startswith[:3] == 'END':
            if acount == 0:
                # If there is no atom record between ENDMDL & END skip to next
                i += 1
                continue
            if model is not None:
                if bonds is None:
                    i += 1
                    break
                else:
                    i += 1
                    for j in range(i, stop):
                        if lines[j].startswith('CONECT'):
                            i = j
                            break
                    continue
            diff = stop - i - 1
            END = diff < acount
            if coordsets is not None:
                END = END or nmodel >= coordsets.shape[0]
            if onlycoords:
                if acount < n_atoms:
                    LOGGER.warn('Discarding model {0}, which contains '
                                '{1} fewer atoms than the first model '
                                'does.'.format(nmodel+1, n_atoms-acount))
                else:
                    coordsets[nmodel] = coordinates
                    nmodel += 1
                acount = 0
                if not END:
                    coordinates = coordsets[nmodel]
            else:
                if acount != n_atoms > 0:
                    raise ValueError('PDB file and AtomGroup ag must have '
                                    'same number of atoms')
                # this is where to decide if more coordsets should be expected
                if END:
                    coordinates.resize((acount, 3), refcheck=False)
                    if addcoords:
                        atomgroup.addCoordset(coordinates, anisous=anisou)
                    else:
                        atomgroup._setCoords(coordinates)
                        if isPDB and anisou is not None:
                            anisou.resize((acount, 6), refcheck=False)
                            atomgroup.setAnisous(anisou / 10000)
                else:
                    coordsets = np.zeros((int(diff//acount+1), acount, 3))
                    coordsets[0] = coordinates[:acount]
                    onlycoords = True
                atomnames.resize(acount, refcheck=False)
                resnames.resize(acount, refcheck=False)
                resnums.resize(acount, refcheck=False)
                chainids.resize(acount, refcheck=False)
                hetero.resize(acount, refcheck=False)
                termini.resize(acount, refcheck=False)
                altlocs.resize(acount, refcheck=False)
                icodes.resize(acount, refcheck=False)
                serials.resize(acount, refcheck=False)
                if not only_subset:
                    atomnames = np.char.strip(atomnames)
                    resnames = np.char.strip(resnames)
                atomgroup.setNames(atomnames)
                atomgroup.setResnames(resnames)
                atomgroup.setResnums(resnums)
                atomgroup.setChids(chainids)
                atomgroup.setFlags('hetatm', hetero)
                atomgroup.setFlags('pdbter', termini)
                atomgroup.setFlags('selpdbter', termini)
                atomgroup.setAltlocs(altlocs)
                if strip_icodes:
                    icodes = np.char.strip(icodes)
                atomgroup.setIcodes(icodes)
                atomgroup.setSerials(serials)
                if isPDB:
                    bfactors.resize(acount, refcheck=False)
                    occupancies.resize(acount, refcheck=False)
                    segnames.resize(acount, refcheck=False)
                    elements.resize(acount, refcheck=False)
                    atomgroup.setBetas(bfactors)
                    atomgroup.setOccupancies(occupancies)
                    atomgroup.setSegnames(np.char.strip(segnames))
                    atomgroup.setElements(np.char.strip(elements))
                    from prody.utilities.misctools import getMasses
                    atomgroup.setMasses(getMasses(np.char.strip(elements)))
                    if anisou is not None:
                        anisou.resize((acount, 6), refcheck=False)
                        atomgroup.setAnisous(anisou / 10000)
                    if siguij is not None:
                        siguij.resize((acount, 6), refcheck=False)
                        atomgroup.setAnistds(siguij / 10000)
                else:
                    radii.resize(acount, refcheck=False)
                    atomgroup.setRadii(radii)
                charges.resize(acount, refcheck=False)
                atomgroup.setCharges(charges)

                nmodel += 1
                n_atoms = acount
                acount = 0
                coordinates = np.zeros((n_atoms, 3), dtype=float)
                if altloc and altloc_torf:
                    _evalAltlocs(atomgroup, altloc, chainids, resnums,
                                 resnames, atomnames)
                    altloc = defaultdict(list)
                if END:
                    break
        elif isPDB and startswith == 'ANISOU':
            if anisou is None:
                anisou = True
                anisou = np.zeros((alength, 6), dtype=float)

            alt = line[16]
            if alt not in which_altlocs and which_altlocs != 'all':
                altloc[alt].append((line, i))
                i += 1
                continue
            try:
                index = acount - 1
                anisou[index, 0] = line[28:35]
                anisou[index, 1] = line[35:42]
                anisou[index, 2] = line[43:49]
                anisou[index, 3] = line[49:56]
                anisou[index, 4] = line[56:63]
                anisou[index, 5] = line[63:70]
            except:
                LOGGER.warn('failed to parse anisotropic temperature '
                    'factors at line {0}'.format(i))
        elif isPDB and startswith =='SIGUIJ':
            if siguij is None:
                siguij = np.zeros((alength, 6),
                    dtype=ATOMIC_FIELDS['siguij'].dtype)
            try:
                index = acount - 1
                siguij[index, 0] = line[28:35]
                siguij[index, 1] = line[35:42]
                siguij[index, 2] = line[43:49]
                siguij[index, 3] = line[49:56]
                siguij[index, 4] = line[56:63]
                siguij[index, 5] = line[63:70]
            except:
                LOGGER.warn('failed to parse standard deviations of '
                    'anisotropic temperature factors at line {0}'.format(i))
        elif startswith =='SIGATM':
            pass
        i += 1
    if onlycoords:
        if acount == atomgroup.numAtoms():
            coordsets[nmodel] = coordinates
            nmodel += 1
        del coordinates
        coordsets.resize((nmodel, atomgroup.numAtoms(), 3), refcheck=False)
        if addcoords:
            atomgroup.addCoordset(coordsets)
        else:
            atomgroup._setCoords(coordsets)
    elif not END:
        # this means last line was an ATOM line, so atomgroup is not decorated
        coordinates.resize((acount, 3), refcheck=False)
        if addcoords:
            atomgroup.addCoordset(coordinates, anisous=anisou)
        else:
            atomgroup._setCoords(coordinates)
        atomnames.resize(acount, refcheck=False)
        resnames.resize(acount, refcheck=False)
        resnums.resize(acount, refcheck=False)
        chainids.resize(acount, refcheck=False)
        hetero.resize(acount, refcheck=False)
        termini.resize(acount, refcheck=False)
        altlocs.resize(acount, refcheck=False)
        icodes.resize(acount, refcheck=False)
        serials.resize(acount, refcheck=False)
        if not only_subset:
            atomnames = np.char.strip(atomnames)
            resnames = np.char.strip(resnames)
        atomgroup.setNames(atomnames)
        atomgroup.setResnames(resnames)
        atomgroup.setResnums(resnums)
        atomgroup.setChids(chainids)
        atomgroup.setFlags('hetatm', hetero)
        atomgroup.setFlags('pdbter', termini)
        atomgroup.setFlags('selpdbter', termini)
        atomgroup.setAltlocs(altlocs)
        if strip_icodes:
            icodes = np.char.strip(icodes)
        atomgroup.setIcodes(icodes)
        atomgroup.setSerials(serials)
        if isPDB:
            if anisou is not None:
                anisou.resize((acount, 6), refcheck=False)
                atomgroup.setAnisous(anisou / 10000)
            if siguij is not None:
                siguij.resize((acount, 6), refcheck=False)
                atomgroup.setAnistds(siguij / 10000)
            bfactors.resize(acount, refcheck=False)
            occupancies.resize(acount, refcheck=False)
            segnames.resize(acount, refcheck=False)
            elements.resize(acount, refcheck=False)
            atomgroup.setSegnames(np.char.strip(segnames))
            atomgroup.setElements(np.char.strip(elements))
            from prody.utilities.misctools import getMasses
            atomgroup.setMasses(getMasses(np.char.strip(elements)))
            atomgroup.setBetas(bfactors)
            atomgroup.setOccupancies(occupancies)
        else:
            radii.resize(acount, refcheck=False)
            atomgroup.setRadii(radii)
        charges.resize(acount, refcheck=False)
        atomgroup.setCharges(charges)

    # rematch the bond-atom serial numbers to atom ids
    serial_to_id = {int(serial): aidx for aidx, serial in enumerate(serials)}
    for bond in bonds_serial:
        try:
            bonds.append([serial_to_id[bond[0]], serial_to_id[bond[1]]])
        except KeyError:
            LOGGER.warn("Bond connecting atom serial numbers {0}"
                        " and {1} contains atoms not included in the model"
                        .format(bond[0], bond[1])
                        )

    if altloc and altloc_torf:
        _evalAltlocs(atomgroup, altloc, chainids, resnums, resnames, atomnames)

    return atomgroup

def _evalAltlocs(atomgroup, altloc, chainids, resnums, resnames, atomnames):
    altloc_keys = list(altloc)
    altloc_keys.sort()
    indices = {}
    for key in altloc_keys:
        xyz = atomgroup.getCoords()
        anisou = atomgroup.getAnisous()
        success = 0
        lines = altloc[key]
        for line, i in lines:
            aan = line[12:16].strip()
            arn = line[17:21].strip()
            ach = line[21]
            ari = int(line[22:26].split()[0])
            rn, ids, ans = indices.get((ach, ari), (None, None, None))
            if ids is None:
                ids = indices.get(ach, None)
                if ids is None:
                    ids = (chainids == ach).nonzero()[0]
                    indices[ach] = ids
                ids = ids[resnums[ids] == ari]
                if len(ids) == 0:
                    LOGGER.warn("failed to parse altloc {0} at line {1}, "
                                "residue not present for altloc 'A'".format(
                                repr(key), i+1))
                    continue
                rn = resnames[ids[0]]
                ans = atomnames[ids]
                indices[(ach, ari)] = (rn, ids, ans)
            if rn != arn:
                LOGGER.warn("failed to parse altloc {0} at line {1}, "
                            "residue name mismatch (expected {2}, "
                            "parsed {3})".format(repr(key), i+1, repr(rn),
                                                 repr(arn)))
                continue
            index = ids[(ans == aan).nonzero()[0]]
            if len(index) != 1:
                LOGGER.warn("failed to parse altloc {0} at line {1}, atom"
                            " {2} not found in the residue"
                            .format(repr(key), i+1, repr(aan)))
                continue

            if line.startswith('ATOM  '):
                try:
                    xyz[index[0], 0] = float(line[30:38])
                    xyz[index[0], 1] = float(line[38:46])
                    xyz[index[0], 2] = float(line[46:54])
                except:
                    LOGGER.warn('failed to parse altloc {0} at line {1}, could'
                                ' not read coordinates'.format(repr(key), i+1))
                    continue
                success += 1
            elif line.startswith('ANISOU'):
                try:
                    anisou[index[0], 0] = line[28:35]
                    anisou[index[0], 1] = line[35:42]
                    anisou[index[0], 2] = line[43:49]
                    anisou[index[0], 3] = line[49:56]
                    anisou[index[0], 4] = line[56:63]
                    anisou[index[0], 5] = line[63:70]
                except:
                    LOGGER.warn('failed to parse altloc {0} at line {1}, could'
                                ' not read anisous'.format(repr(key), i+1))
                success += 1
        LOGGER.info('{0} out of {1} altloc {2} lines were parsed.'
                    .format(success, len(lines), repr(key)))
        if success > 0:
            LOGGER.info('Altloc {0} is appended as a coordinate set to '
                        'atomgroup {1}.'.format(repr(key), atomgroup.getTitle()))
            atomgroup.addCoordset(xyz, label='altloc ' + key, anisous=anisou)


#HELIXLINE = ('HELIX  %3d %3s %-3s %1s %4d%1s %-3s %1s %4d%1s%2d'
#             '                               %5d\n')

HELIXLINE = ('HELIX  {serNum:3d} {helixID:>3s} '
             '{initResName:<3s}{initChainID:>2s} {initSeqNum:4d}{initICode:1s} '
             '{endResName:<3s}{endChainID:>2s} {endSeqNum:4d}{endICode:1s}'
             '{helixClass:2d}                               {length:5d}    \n')

SHEETLINE = ('SHEET  {strand:3d} {sheetID:>3s}{numStrands:2d} '
             '{initResName:3s}{initChainID:>2s}{initSeqNum:4d}{initICode:1s} '
             '{endResName:3s}{endChainID:>2s}{endSeqNum:4d}{endICode:1s}{sense:2d} \n')

PDBLINE_LT100K = ('%-6s%5d %-4s%1s%-3s%2s%4d%1s   '
                  '%8.3f%8.3f%8.3f%6.2f%6.2f      '
                  '%4s%2s%2s\n')

# Residue number
PDBLINE_GE10K = ('%-6s%5d %-4s%1s%-3s%2s%4x%1s   '
                 '%8.3f%8.3f%8.3f%6.2f%6.2f      '
                 '%4s%2s%2s\n')

# Serial number
PDBLINE_GE100K = ('%-6s%5x %-4s%1s%-3s%2s%4d%1s   '
                  '%8.3f%8.3f%8.3f%6.2f%6.2f      '
                  '%4s%2s%2s\n')

# Both
PDBLINE_GE100K_GE10K = ('%-6s%5x %-4s%1s%-3s%2s%4x%1s   '
                        '%8.3f%8.3f%8.3f%6.2f%6.2f      '
                        '%4s%2s%2s\n')

# All cases
PDBLINE_H36 = ('%-6s%5s %-4s%1s%-3s%2s%4s%1s   '
               '%8.3f%8.3f%8.3f%6.2f%6.2f      '
               '%4s%2s%2s\n')

ANISOULINE_LT100K = ('%-6s%5d %-4s%1s%-3s%2s%4d%1s '
                     '%7d%7d%7d%7d%7d%7d  '
                     '%4s%2s%2s\n')

# Residue number
ANISOULINE_GE10K = ('%-6s%5d %-4s%1s%-3s%2s%4x%1s '
                    '%7d%7d%7d%7d%7d%7d  '
                    '%4s%2s%2s\n')

# Serial number
ANISOULINE_GE100K = ('%-6s%5x %-4s%1s%-3s%2s%4d%1s '
                     '%7d%7d%7d%7d%7d%7d  '
                     '%4s%2s%2s\n')

# Both
ANISOULINE_GE100K_GE10K = ('%-6s%5x %-4s%1s%-3s%2s%4x%1s '
                           '%7d%7d%7d%7d%7d%7d  '
                           '%4s%2s%2s\n')

# All cases
ANISOULINE_H36 = ('%-6s%5s %-4s%1s%-3s%2s%4s%1s '
                  '%7d%7d%7d%7d%7d%7d  '
                  '%4s%2s%2s\n')

_writePDBdoc = """

    :arg atoms: an object with atom and coordinate data

    :arg csets: coordinate set indices, default is all coordinate sets

    :arg beta: a list or array of number to be outputted in beta column

    :arg occupancy: a list or array of number to be outputted in occupancy
        column

    :arg hybrid36: whether to use hybrid36 format for atoms with serial
        greater than 99999. Hexadecimal is used otherwise.
        Default is False
    :type hybrid36: bool 
    """

def writeChainsList(chains, filename):
    """
    Write a text file containing a list of chains that can be parsed.

    :arg chains: a list of :class:`.Chain` objects
    :type chains: list

    :arg filename: the name of the file to be written
    :type filename: str
    """

    fo = open(filename,'w')
    for chain in chains:
        fo.write(chain.getTitle() + ' ' + chain.getChid() + '\n')
    fo.close()
    return

def parseChainsList(filename):
    """
    Parse a set of PDBs and extract chains based on a list in a text file.

    :arg filename: the name of the file to be read
    :type filename: str

    Returns: lists containing an :class:'.AtomGroup' for each PDB, 
    the headers for those PDBs, and the requested :class:`.Chain` objects
    """

    fi = open(filename,'r')
    lines = fi.readlines()
    fi.close()

    pdb_ids = []
    ags = []
    headers = []
    chains = []
    num_lines = len(lines)
    LOGGER.progress('Starting', num_lines, '_prody_parseChainsList')
    for i, line in enumerate(lines):
        LOGGER.update(i, 'Parsing lines...', label='_prody_parseChainsList')
        pdb_id = line.split()[0].split('_')[0]
        if not pdb_id in pdb_ids:
            pdb_ids.append(pdb_id)

            ag, header = parsePDB(pdb_id, compressed=False, \
                                  subset=line.split()[0].split('_')[1], header=True)

            ags.append(ag)
            headers.append(header)

        chains.append(ag.getHierView()[line.strip().split()[1]])

    LOGGER.finish()
    LOGGER.info('{0} PDBs have been parsed and {1} chains have been extracted. \
                '.format(len(ags),len(chains)))

    return ags, headers, chains

def writePDBStream(stream, atoms, csets=None, **kwargs):
    """Write *atoms* in PDB format to a *stream*.

    :arg stream: anything that implements a :meth:`write` method (e.g. file,
        buffer, stdout)

    :arg renumber: whether to renumber atoms with serial indices
        Default is **True**
    :type renumber: bool

    :arg hybrid36: whether to use hybrid36 format for atom residue numbers
        Default is **False**, which means using hexadecimal instead.
        NB: ChimeraX seems to prefer hybrid36 and may have problems with hexadecimal.
    :type hybrid36: bool

    :arg full_ter: whether to write full TER lines with atoms info
        Default is **True**
    :type full_ter: bool

    :arg write_remarks: whether to write REMARK lines
        Default is **True**
    :type write_remarks: bool
    """    
    renumber = kwargs.get('renumber', True)
    full_ter = kwargs.get('full_ter', True)
    write_remarks = kwargs.get('write_remarks', True)
    long_resname = kwargs.get('long_resname', False)

    remark = str(atoms)
    try:
        coordsets = atoms._getCoordsets(csets)
    except AttributeError:
        try:
            coordsets = atoms._getCoords()
        except AttributeError:
            raise TypeError('atoms must be an object with coordinate sets')
        if coordsets is not None:
            coordsets = [coordsets]
    else:
        if coordsets.ndim == 2:
            coordsets = [coordsets]
    if coordsets is None:
        raise ValueError('atoms does not have any coordinate sets')

    had_atoms = False
    try:
        acsi = atoms.getACSIndex()
        had_atoms = True
    except AttributeError:
        try:
            atoms = atoms.getAtoms()
        except AttributeError:
            raise TypeError('atoms must be an Atomic instance or an object '
                            'with `getAtoms` method')
        else:
            if atoms is None:
                raise ValueError('atoms is not associated with an Atomic '
                                 'instance')
            try:
                acsi = atoms.getACSIndex()
            except AttributeError:
                raise TypeError('atoms does not have a valid type')

    try:
        atoms.getIndex()
    except AttributeError:
        pass
    else:
        atoms = atoms.select('all')

    n_atoms = atoms.numAtoms()

    hybrid36 = kwargs.get('hybrid36', False)

    occupancy = kwargs.get('occupancy')
    if occupancy is None:
        occupancies = atoms._getOccupancies()
        if occupancies is None:
            occupancies = np.zeros(n_atoms, float)
    else:
        occupancies = np.array(occupancy)
        if len(occupancies) != n_atoms:
            raise ValueError('len(occupancy) must be equal to number of atoms')

    beta = kwargs.get('beta')
    if beta is None:
        bfactors = atoms._getBetas()
        if bfactors is None:
            bfactors = np.zeros(n_atoms, float)
    else:
        bfactors = np.array(beta)
        if len(bfactors) != n_atoms:
            raise ValueError('len(beta) must be equal to number of atoms')

    atomnames = atoms.getNames()
    if atomnames is None:
        raise ValueError('atom names are not set')
    for i, an in enumerate(atomnames):
        if len(an) < 4:
            atomnames[i] = ' ' + an

    s_or_u = np.array(['a']).dtype.char

    altlocs = atoms._getAltlocs()
    if altlocs is None:
        altlocs = np.zeros(n_atoms, s_or_u + '1')

    resnames = atoms._getResnames()
    if resnames is None:
        resnames = ['UNK'] * n_atoms

    chainids = atoms._getChids()
    if chainids is None:
        chainids = np.zeros(n_atoms, s_or_u + '1')

    resnums = atoms._getResnums()
    if resnums is None:
        resnums = np.ones(n_atoms, int)

    serials = atoms._getSerials()
    if serials is None or renumber:
        serials = np.arange(n_atoms, dtype=int) + 1

    icodes = atoms._getIcodes()
    if icodes is None:
        icodes = np.zeros(n_atoms, s_or_u + '1')

    hetero = ['ATOM'] * n_atoms
    heteroflags = atoms._getFlags('hetatm')
    if heteroflags is None:
        heteroflags = atoms._getFlags('hetero')
    if heteroflags is not None:
        hetero = np.array(hetero, s_or_u + '6')
        hetero[heteroflags] = 'HETATM'

    elements = atoms._getElements()
    if elements is None:
        elements = np.zeros(n_atoms, s_or_u + '1')
    else:
        elements = np.char.rjust(elements, 2)

    segments = atoms._getSegnames()
    if segments is None:
        segments = np.zeros(n_atoms, s_or_u + '6')

    charges = atoms._getCharges()
    charges2 = np.empty(n_atoms, s_or_u + '2')
    if charges is not None:
        for i, charge in enumerate(charges):
            charges2[i] = str(abs(int(charge)))

            if np.sign(charge) == -1:
                charges2[i] += '-'
            else:
                charges2[i] += '+'

            if charges2[i] == '0+':
                charges2[i] = '  '

    if write_remarks:
        # write remarks
        stream.write('REMARK {0}\n'.format(remark))

        # write secondary structures (if any)
        secondary = kwargs.get('secondary', True)
        secstrs = atoms._getSecstrs()
        if secstrs is not None and secondary:
            secindices = atoms._getSecindices()
            secclasses = atoms._getSecclasses()
            secids = atoms._getSecids()

            # write helices
            for i in range(1,max(secindices)+1):
                torf = np.logical_and(isHelix(secstrs), secindices==i)
                if torf.any():
                    helix_resnums = resnums[torf]
                    helix_chainids = chainids[torf]
                    helix_resnames = resnames[torf]
                    helix_secclasses = secclasses[torf]
                    helix_secids = secids[torf]
                    helix_icodes = icodes[torf]
                    L = helix_resnums[-1] - helix_resnums[0] + 1

                    stream.write(HELIXLINE.format(serNum=i, helixID=helix_secids[0], 
                                initResName=helix_resnames[0][:3], initChainID=helix_chainids[0], 
                                initSeqNum=helix_resnums[0], initICode=helix_icodes[0],
                                endResName=helix_resnames[-1][:3], endChainID=helix_chainids[-1], 
                                endSeqNum=helix_resnums[-1], endICode=helix_icodes[-1],
                                helixClass=helix_secclasses[0], length=L))

            # write strands
            torf_all_sheets = isSheet(secstrs)
            sheet_secids = secids[torf_all_sheets]

            unique_sheet_secids, indices = np.unique(sheet_secids, return_index=True)
            unique_sheet_secids = unique_sheet_secids[indices.argsort()]
            for sheet_id in unique_sheet_secids:
                torf_strands_in_sheet = np.logical_and(torf_all_sheets, secids==sheet_id)
                strand_indices = secindices[torf_strands_in_sheet]
                numStrands = len(np.unique(strand_indices))

                for i in np.unique(strand_indices):
                    torf_strand = np.logical_and(torf_strands_in_sheet, secindices==i)
                    strand_resnums = resnums[torf_strand]
                    strand_chainids = chainids[torf_strand]
                    strand_resnames = resnames[torf_strand]
                    strand_secclasses = secclasses[torf_strand]
                    strand_icodes = icodes[torf_strand]

                    stream.write(SHEETLINE.format(strand=i, sheetID=sheet_id, numStrands=numStrands,
                                initResName=strand_resnames[0][:3], initChainID=strand_chainids[0], 
                                initSeqNum=strand_resnums[0], initICode=strand_icodes[0],
                                endResName=strand_resnames[-1][:3], endChainID=strand_chainids[-1], 
                                endSeqNum=strand_resnums[-1], endICode=strand_icodes[-1],
                                sense=strand_secclasses[0]))
            pass

    # write atoms
    multi = len(coordsets) > 1
    write = stream.write
    num_ter_lines = 0
    for m, coords in enumerate(coordsets):

        if had_atoms:
            atoms.setACSIndex(m)
        anisous = atoms._getAnisous()
        if anisous is not None:
            anisous = np.array(anisous * 10000, dtype=int)

        if multi:
            write('MODEL{0:9d}\n'.format(m+1))

        if not hybrid36:
            # We need to check whether serial and residue numbers become hexadecimal
            reached_max_n_atom = False
            reached_max_n_res = False

            pdbline = PDBLINE_LT100K
            anisouline = ANISOULINE_LT100K
        else:
            warned_hybrid36 = False

        warned_5_digit = False
        warned_long_resname = False

        for i, xyz in enumerate(coords):

            serial = serials[i] + num_ter_lines

            if hybrid36:
                pdbline = PDBLINE_H36
                anisouline = ANISOULINE_H36

                if not warned_hybrid36:
                    LOGGER.warn('hybrid36 format is being used')
                    warned_hybrid36 = True

            else:
                if not (reached_max_n_atom or reached_max_n_res) and (i == MAX_N_ATOM or serial > MAX_N_ATOM):
                    reached_max_n_atom = True
                    pdbline = PDBLINE_GE100K
                    anisouline = ANISOULINE_GE100K
                    LOGGER.warn('Indices are exceeding 99999 and hexadecimal format is being used for indices')

                elif not (reached_max_n_atom or reached_max_n_res) and resnums[i] > MAX_N_RES:
                    reached_max_n_res = True
                    pdbline = PDBLINE_GE10K
                    anisouline = ANISOULINE_GE10K
                    LOGGER.warn('Resnums are exceeding 9999 and hexadecimal format is being used for resnums')

                elif reached_max_n_atom and not reached_max_n_res and resnums[i] > MAX_N_RES:
                    reached_max_n_res = True
                    pdbline = PDBLINE_GE100K_GE10K
                    anisouline = ANISOULINE_GE100K_GE10K
                    LOGGER.warn('Resnums are exceeding 9999 and hexadecimal format is being used for indices and resnums')

                elif reached_max_n_res and not reached_max_n_atom and (i == MAX_N_ATOM or serial > MAX_N_ATOM):
                    reached_max_n_atom = True
                    pdbline = PDBLINE_GE100K_GE10K
                    anisouline = ANISOULINE_GE100K_GE10K
                    LOGGER.warn('Indices are exceeding 99999 and hexadecimal format is being used for indices and resnums')

            if hybrid36:
                serial = decToHybrid36(serial)
                resnum = decToHybrid36(resnums[i], resnum=True)
            else:
                resnum = resnums[i]

            resname = resnames[i]
            resname_len = 3
            if long_resname:
                resname_len = 4
                pdbline = pdbline.replace('%-3s%2s', '%-4s%1s')

            if len(resnames[i]) > resname_len:
                resname = resname[:resname_len]
                if not warned_long_resname:
                    LOGGER.warn('Resname {0} too long, cutting resname to {1} characters as {2}'.format(
                        resnames[i], resname_len, resname
                    ))
                    warned_long_resname = True

            if pdbline == PDBLINE_LT100K or hybrid36:
                if len(str(resnum)) == 5:
                    if icodes[i] == '':
                        icodes[i] = str(resnum)[4]

                        if not warned_5_digit:
                            LOGGER.warn('Storing 5-digit resnums using insertion codes')
                            warned_5_digit = True
                    else:
                        LOGGER.warn('Truncating 5-digit resnum as insertion code is busy.')

                    resnum = int(str(resnum)[:4])

                elif len(str(resnum)) > 5:
                    if not warned_5_digit:
                        LOGGER.warn('Truncating {0}-digit resnum as too long to be '
                                    'supported by insertion code.'.format(len(str(resnum))))
                        warned_5_digit = True

                    resnum = int(str(resnum)[:4])
            else:
                final_resnum = '%4x' % int(resnum)

                if len(str(final_resnum)) == 5:
                    if icodes[i] == '':
                        icodes[i] = str(final_resnum)[4]

                        if not warned_5_digit:
                            LOGGER.warn('Storing 5-digit hex resnums using insertion codes')
                            warned_5_digit = True
                    else:
                        LOGGER.warn('Truncating 5-digit hex resnum as insertion code is busy.')

                    resnum = int(str(final_resnum)[:4], 16)

                elif len(str(final_resnum)) > 5:
                    if not warned_5_digit:
                        LOGGER.warn('Truncating {0}-digit hex resnum ({1}) as too long to be '
                                    'supported by insertion code.'.format(len(str(final_resnum)), 
                                                                          final_resnum))
                        warned_5_digit = True

                    resnum = int(str(final_resnum)[:4], 16)


            write(pdbline % (hetero[i], serial,
                             atomnames[i], altlocs[i],
                             resname, chainids[i][:1], resnum,
                             icodes[i],
                             xyz[0], xyz[1], xyz[2],
                             occupancies[i], bfactors[i],
                             segments[i], elements[i], charges2[i]))

            if anisous is not None:
                anisou = anisous[i]

                write(anisouline % ("ANISOU", serial,
                                    atomnames[i], altlocs[i],
                                    resname, chainids[i], resnum,
                                    icodes[i],
                                    anisou[0], anisou[1], anisou[2],
                                    anisou[3], anisou[4], anisou[5],
                                    segments[i], elements[i], charges2[i]))

            if isinstance(atoms, AtomGroup):
                if resnames[i] in AAMAP and atoms._getFlags('pdbter') is not None and atoms.getFlags('pdbter')[i]:
                    if hybrid36:
                        serial = decToHybrid36(hybrid36ToDec(serial) + 1)
                    else:
                        serial += 1

                    if full_ter:
                        false_pdbline = pdbline % ("TER   ", serial,
                                                   "", "",
                                                   resname, chainids[i], resnum,
                                                   icodes[i],
                                                   xyz[0], xyz[1], xyz[2],
                                                   occupancies[i], bfactors[i],
                                                   segments[i], elements[i], charges2[i])
                    else:
                        false_pdbline = "TER" + " "*23
                    write(false_pdbline[:26] + " "*54 + '\n')
                    num_ter_lines += 1
            else:
                if resnames[i] in AAMAP and atoms._getFlags('selpdbter') is not None and atoms.getFlags('selpdbter')[i]:
                    if hybrid36:
                        serial = decToHybrid36(hybrid36ToDec(serial) + 1)
                    else:
                        serial += 1

                    if full_ter:
                        false_pdbline = pdbline % ("TER   ", serial,
                                                   "", "",
                                                   resname, chainids[i], resnum,
                                                   icodes[i],
                                                   xyz[0], xyz[1], xyz[2],
                                                   occupancies[i], bfactors[i],
                                                   segments[i], elements[i], charges2[i])
                    else:
                        false_pdbline = "TER" + " "*23
                    write(false_pdbline[:26] + " "*54 + '\n')
                    num_ter_lines += 1

        if multi:
            write('ENDMDL' + " "*74 + '\n')
            altlocs = np.zeros(n_atoms, s_or_u + '1')
            
    write('END   ' + " "*74 + '\n')

    if had_atoms:
        atoms.setACSIndex(acsi)

writePDBStream.__doc__ += _writePDBdoc

def writePDB(filename, atoms, csets=None, autoext=True, **kwargs):
    """Write *atoms* in PDB format to a file with name *filename* and return
    *filename*.  If *filename* ends with :file:`.gz`, a compressed file will
    be written.

    :arg renumber: whether to renumber atoms with serial indices
        Default is **True**
    :type renumber: bool

    :arg hybrid36: whether to use hybrid36 format for atom residue numbers
        Default is **False**, which means using hexadecimal instead.
        NB: ChimeraX seems to prefer hybrid36 and may have problems with hexadecimal.
    :type hybrid36: bool

    Please note that resnames longer than 3 characters will be trimmed by default.
    This behaviour can be changed with the following argument:

    :arg long_resname: whether to write 4-character resnames instead of cutting at 3 chars
    :type long_resname: bool
    """

    if not (filename.lower().endswith('.pdb') or filename.lower().endswith('.pdb.gz') or
            filename.lower().endswith('.ent') or filename.lower().endswith('.ent.gz')) and autoext:
        filename += '.pdb'
    out = openFile(filename, 'wt')
    writePDBStream(out, atoms, csets, **kwargs)
    out.close()
    return filename

writePDB.__doc__ += _writePDBdoc + """
    :arg autoext: when not present, append extension :file:`.pdb` to *filename*
"""

def writePQRStream(stream, atoms, **kwargs):
    if isinstance(atoms, Atom):
        atoms = Selection(atoms.getAtomGroup(), [atoms.getIndex()],
                          atoms.getACSIndex(),
                          'index ' + str(atoms.getIndex()))
    n_atoms = atoms.numAtoms()
    atomnames = atoms.getNames()
    if atomnames is None:
        raise RuntimeError('atom names are not set')
    for i, an in enumerate(atomnames):
        lenan = len(an)
        if lenan < 4:
            atomnames[i] = ' ' + an
        elif lenan > 4:
            atomnames[i] = an[:4]

    s_or_u = np.array(['a']).dtype.char


    write = stream.write

    calphas = atoms.ca
    if calphas is not None:
        ssa = calphas.getSecstrs()
    else:
        ssa = None
    helix = []
    sheet = []
    if ssa is not None:
        ss_prev = ssa[0]
        ss_start = 0
        ss_end = 1
        for i, ss in enumerate(ssa):
            if ss != ss_prev:
                # store prev secstr and prepare for next
                ss_end = i-1
                init = calphas[ss_start]
                end = calphas[ss_end]
                length = ss_end - ss_start + 1

                entries = [init.getSecindex(), init.getSecid(),
                           init.getResname(), init.getChid(), 
                           init.getResnum(), init.getIcode(),
                           end.getResname(), end.getChid(),
                           end.getResnum(), end.getIcode(),
                           init.getSecclass()]

                if ssa[ss_end] == 'H':
                    helix.append(["HELIX "] + entries +
                                 ['', length])

                elif ssa[ss_end] == 'E':
                    sheet.append(["SHEET "] + entries)

                ss_start = i
                ss_prev = ss

    format_helix = ('{0:6s} {1:3d} {2:3s} ' +
                    '{3:3s} {4:1s} {5:4d}{6:1s} ' +
                    '{7:3s} {8:1s} {9:4d}{10:1s} ' +
                    '{11:2d} {12:30s} {13:5d}\n').format
    for line in helix:
        write(format_helix(*line))


    sheet_prev = 'A'
    num_strands_list = []
    for i, item1 in enumerate(sheet):
        if item1[2] != sheet_prev:
            num_strands = sheet[i-1][1]
            num_strands_list.append(num_strands)

            sheet_prev = item1[2]

            for item2 in sheet[i-num_strands:i]:
                item2.append(num_strands)

    num_strands = len(num_strands_list)
    for item2 in sheet[i-num_strands+1:]:
        item2.append(num_strands)

    format_sheet = ('{0:6s} {1:3d} {2:3s}{12:2d} ' +
                    '{3:3s} {4:1s}{5:4d}{6:1s}' +
                    '{7:3s} {8:1s}{9:4d}{10:1s}' +
                    '{11:2d}\n').format

    for i, line in enumerate(sheet):
        write(format_sheet(*line))

    resnames = atoms._getResnames()
    if resnames is None:
        resnames = ['UNK'] * n_atoms
    resnums = atoms._getResnums()
    if resnums is None:
        resnums = np.ones(n_atoms, int)
    chainids = atoms._getChids()
    if chainids is None:
        chainids = np.zeros(n_atoms, s_or_u + '1')
    charges = atoms._getCharges()
    if charges is None:
        charges = np.zeros(n_atoms, float)
    radii = atoms._getRadii()
    if radii is None:
        radii = np.zeros(n_atoms, float)
    icodes = atoms._getIcodes()
    if icodes is None:
        icodes = np.zeros(n_atoms, s_or_u + '1')
    hetero = ['ATOM'] * n_atoms
    heteroflags = atoms._getFlags('hetatm')
    if heteroflags is None:
        heteroflags = atoms._getFlags('hetero')
    if heteroflags is not None:
        hetero = np.array(hetero, s_or_u + '6')
        hetero[heteroflags] = 'HETATM'
    altlocs = atoms._getAltlocs()
    if altlocs is None:
        altlocs = np.zeros(n_atoms, s_or_u + '1')

    format = ('{0:6s} {1:5d} {2:4s} {3:1s}' +
              '{4:4s} {5:1s} {6:4d} {7:1s}   ' +
              '{8:8.3f} {9:8.3f} {10:8.3f}' +
              '{11:8.4f} {12:7.4f}\n').format
    coords = atoms._getCoords()

    for i, xyz in enumerate(coords):
        write(format(hetero[i], i+1, atomnames[i], altlocs[i],
                     resnames[i], chainids[i], int(resnums[i]),
                     icodes[i], xyz[0], xyz[1], xyz[2], charges[i], radii[i]))

def writePQR(filename, atoms, **kwargs):
    """Write *atoms* in PQR format to a file with name *filename*.  Only
    current coordinate set is written.  Returns *filename* upon success.  If
    *filename* ends with :file:`.gz`, a compressed file will be written."""

    stream = openFile(filename, 'w')
    writePQRStream(stream, atoms, **kwargs)
    stream.close()
    return filename



# -*- coding: utf-8 -*-
"""This module defines functions for generating alternate conformations along
normal modes."""

import numpy as np

from prody import LOGGER
from prody.atomic import Atomic, AtomGroup
from prody.ensemble import Ensemble

from .nma import NMA
from .mode import Mode, VectorBase
from .modeset import ModeSet

__all__ = ['deformAtoms', 'sampleModes', 'traverseMode']


def sampleModes(modes, atoms=None, n_confs=1000, rmsd=1.0):
    """Returns an ensemble of randomly sampled conformations along given
    *modes*.  If *atoms* are provided, sampling will be around its active
    coordinate set.  Otherwise, sampling is around the 0 coordinate set.

    :arg modes: modes along which sampling will be performed
    :type modes: :class:`.Mode`, :class:`.ModeSet`, :class:`.PCA`,
                 :class:`.ANM` or :class:`.NMA`

    :arg atoms: atoms whose active coordinate set will be used as the initial
        conformation
    :type atoms: :class:`.Atomic`

    :arg n_confs: number of conformations to generate, default is 1000
    :type n_steps: int

    :arg rmsd: average RMSD that the conformations will have with
        respect to the initial conformation, default is 1.0 Å
    :type rmsd: float

    :returns: :class:`.Ensemble`

    For given normal modes :math:`[u_1 u_2 ... u_m]` and their eigenvalues
    :math:`[\\lambda_1 \\lambda_2 ... \\lambda_m]`, a new conformation
    is sampled using the relation:

    .. math::

       R_k = R_0 + s \\sum_{i=1}^{m} r_i^k \\lambda^{-0.5}_i u_i

    :math:`R_0` is the active coordinate set of *atoms*.
    :math:`[r_1^k r_2^k ... r_m^k]` are normally distributed random numbers
    generated for conformation :math:`k` using :func:`numpy.random.randn`.

    RMSD of the new conformation from :math:`R_0` can be calculated as


    .. math::

      RMSD^k = \\sqrt{  \\left[ s \\sum_{i=1}^{m} r_i^k \\lambda^{-0.5}_i u_i \\right] ^{2} / N } = \\frac{s}{ \\sqrt{N}} \\sqrt{ \\sum_{i=1}^{m} (r_i^k)^2 \\lambda^{-1}_i  }


    Average :math:`RMSD` of the generated conformations from the initial conformation is:

    .. math::

      \\left< RMSD^k \\right> = \\frac{s}{ \\sqrt{N}} \\left< \\sqrt{ \\sum_{i=1}^{m} (r_i^k)^2 \\lambda^{-1}_i } \\right>


    From this relation :math:`s` scaling factor obtained using the relation

    .. math::

       s =  \\left< RMSD^k \\right> \\sqrt{N} {\\left< \\sqrt{ \\sum_{i=1}^{m} (r_i)^2 \\lambda^{-1}_i} \\right>}^{-1}


    Note that random numbers are generated before conformations are
    sampled, hence exact value of :math:`s` is known from this relation to
    ensure that the generated ensemble will have user given average *rmsd*
    value.

    Note that if modes are from a :class:`.PCA`, variances are used instead of
    inverse eigenvalues, i.e. :math:`\\sigma_i \\sim \\lambda^{-1}_i`.

    See also :func:`.showEllipsoid`."""

    if not isinstance(modes, (Mode, NMA, ModeSet)):
        raise TypeError('modes must be a NMA or ModeSet instance, '
                        'not {0}'.format(type(modes)))
    if not modes.is3d():
        raise ValueError('modes must be from a 3-dimensional model')
    n_confs = int(n_confs)
    n_atoms = modes.numAtoms()
    initial = None
    if atoms is not None:
        if isinstance(atoms, Atomic):
            if atoms.numAtoms() != n_atoms:
                raise ValueError('number of atoms do not match')
            initial = atoms.getCoords()
        elif isinstance(atoms, np.ndarray):
            initial = atoms
        else:
            raise TypeError('{0} is not correct type for atoms'
                            .format(type(atoms)))

    rmsd = float(rmsd)
    LOGGER.info('Parameter: rmsd = {0:.2f} A'.format(rmsd))
    n_confs = int(n_confs)
    LOGGER.info('Parameter: n_confs = {0}'.format(n_confs))

    if isinstance(modes, Mode):
        n_modes = 1
        variances = np.array([modes.getVariance()])
        magnitudes = np.array([abs(modes)])
    else:
        n_modes = len(modes)
        variances = modes.getVariances()
        magnitudes = np.array([abs(mode) for mode in modes])

    if np.any(variances == 0):
        raise ValueError('one or more modes has zero variance')
    randn = np.random.standard_normal((n_confs, n_modes))
    coef = ((randn ** 2 * variances).sum(1) ** 0.5).mean()
    scale = n_atoms**0.5 * rmsd / coef

    LOGGER.info('Modes are scaled by {0}.'.format(scale))

    confs = []
    append = confs.append
    scale = scale / magnitudes * variances ** 0.5

    array = modes._getArray()
    if array.ndim > 1:
        for i in range(n_confs):
            append((array * scale * randn[i]).sum(1).reshape((n_atoms, 3)))
    else:
        for i in range(n_confs):
            append((array * scale * randn[i]).reshape((n_atoms, 3)))

    ensemble = Ensemble('Conformations along {0}'.format(modes))
    if initial is None:
        ensemble.setCoords(np.zeros((n_atoms, 3)))
        ensemble.addCoordset(np.array(confs))
    else:
        ensemble.setCoords(initial)
        ensemble.addCoordset(np.array(confs) + initial)
    return ensemble


def traverseMode(mode, atoms, n_steps=10, rmsd=1.5, **kwargs):
    """Generates a trajectory along a given *mode*, which can be used to
    animate fluctuations in an external program.

    :arg mode: mode along which a trajectory will be generated
    :type mode: :class:`.Mode`

    :arg atoms: atoms whose active coordinate set will be used as the initial
        conformation
    :type atoms: :class:`.Atomic`

    :arg n_steps: number of steps to take along each direction,
        for example, for ``n_steps=10``, 20 conformations will be
        generated along *mode* with structure *atoms* in between, 
        default is 10.
    :type n_steps: int

    :arg rmsd: maximum RMSD that the conformations will have with
        respect to the initial conformation, default is 1.5 Å
    :type rmsd: float

    :arg pos: whether to include steps in the positive mode
        direction, default is **True**
    :type pos: bool

    :arg neg: whether to include steps in the negative mode
        direction, default is **True**
    :type neg: bool

    :arg reverse: whether to reverse the direction
        default is **False**
    :type reverse: bool

    :returns: :class:`.Ensemble`

    For given normal mode :math:`u_i`, its eigenvalue
    :math:`\\lambda_i`, number of steps :math:`n`, and maximum :math:`RMSD`
    conformations :math:`[R_{-n} R_{-n+1} ... R_{-1} R_0 R_1 ... R_n]` are
    generated.

    :math:`R_0` is the active coordinate set of *atoms*.
    :math:`R_k = R_0 + sk\\lambda_iu_i`, where :math:`s` is found using
    :math:`s = ((N (\\frac{RMSD}{n})^2) / \\lambda_i^{-1}) ^{0.5}`, where
    :math:`N` is the number of atoms."""

    pos = kwargs.get('pos', True)
    neg = kwargs.get('neg', True)
    reverse = kwargs.get('reverse', False)

    if pos is False and neg is False:
        raise ValueError('pos and neg cannot both be False')

    if not isinstance(mode, VectorBase):
        raise TypeError('mode must be a Mode or Vector instance, '
                        'not {0}'.format(type(mode)))
    if not mode.is3d():
        raise ValueError('mode must be from a 3-dimensional model.')
    n_atoms = mode.numAtoms()
    initial = None
    if atoms is not None:
        if not isinstance(atoms, Atomic):
            raise TypeError('{0} is not correct type for atoms'
                            .format(type(atoms)))
        if atoms.numAtoms() != n_atoms:
            raise ValueError('number of atoms do not match')
        initial = atoms.getCoords()

    name = str(mode)

    rmsd = float(rmsd) + 0.000004
    LOGGER.info('Parameter: rmsd = {0:.2f} A'.format(rmsd))
    n_steps = int(n_steps)
    LOGGER.info('Parameter: n_steps = {0}'.format(n_steps))
    step = rmsd / n_steps
    LOGGER.info('Step size is {0:.2f} A RMSD'.format(step))
    arr = mode.getArrayNx3()
    try:
        var = mode.getVariance()
    except AttributeError:
        var = 1.
    scale = ((n_atoms * step**2) / var) ** 0.5
    LOGGER.info('Mode is scaled by {0}.'.format(scale))

    array = arr * var**0.5 * scale / abs(mode)
    confs_add = [initial + array]
    for s in range(1, n_steps):
        confs_add.append(confs_add[-1] + array)
    confs_sub = [initial - array]
    for s in range(1, n_steps):
        confs_sub.append(confs_sub[-1] - array)
    confs_sub.reverse()
    ensemble = Ensemble('Conformations along {0}'.format(name))
    ensemble.setAtoms(atoms)
    ensemble.setCoords(initial)

    conf_list = [initial]
    if pos:
        conf_list = conf_list + confs_add
    if  neg:
        conf_list = confs_sub + conf_list
    conf_array = np.array(conf_list)

    if reverse:
        conf_array = conf_array[::-1]

    ensemble.addCoordset(conf_array)
    return ensemble


def deformAtoms(atoms, mode, rmsd=None, replace=False, scale=None):
    """Generate a new coordinate set for *atoms* along the *mode*.  *atoms*
    must be a :class:`.AtomGroup` instance.  New coordinate set will be
    appended to *atoms*. If *rmsd* is provided, *mode* will be scaled to
    generate a coordinate set with given RMSD distance to the active coordinate
    set."""

    if not isinstance(atoms, Atomic):
        raise TypeError('atoms must be an Atomic object, not {0}'
                        .format(type(atoms)))
    if not isinstance(mode, VectorBase):
        raise TypeError('mode must be a Mode or Vector instance, '
                        'not {0}'.format(type(mode)))
    if not mode.is3d():
        raise ValueError('mode must be from a 3-dimensional model.')
    if atoms.numAtoms() != mode.numAtoms():
        raise ValueError('number of atoms do not match')

    if scale is None: 
        scale = 1.

    array = mode.getArrayNx3()

    if rmsd is not None:
        rmsd = float(rmsd)
        # rmsd = ( ((scalar * array)**2).sum() / n_atoms )**0.5
        scalar = (atoms.numAtoms() * rmsd**2 / (array**2).sum())**0.5
        scale *= scalar
        LOGGER.info('Mode is scaled by {0}.'.format(scalar))

    if replace is False:
        atoms.addCoordset(atoms.getCoords() + array * scale)
    else:
        atoms.setCoords(atoms.getCoords() + array * scale)





"""This module contains unit tests for :mod:`~prody.ensemble`."""

from prody.tests import TestCase

from numpy.testing import assert_equal

from prody import (calcOccupancies, trimPDBEnsemble, PDBEnsemble, 
                   parsePDB, buildPDBEnsemble, bestMatch, sameChid, 
                   sameChainPos)
from prody.tests.datafiles import *
from . import PDBENSEMBLE, WEIGHTS, ENSEMBLE, ATOMS, PDBENSEMBLEA


class TestCalcOccupancies(TestCase):

    def testResults(self):
        assert_equal(calcOccupancies(PDBENSEMBLE), WEIGHTS.sum(0).flatten(),
                     'calcOccupancies failed')

    def testInvalidType(self):

        self.assertRaises(TypeError, calcOccupancies, ENSEMBLE)

    def testWeightsNone(self):

        self.assertRaises(ValueError, calcOccupancies, PDBEnsemble())

class TestTrimPDBEnsemble(TestCase):

    def testResults(self):
        ensemble = PDBENSEMBLEA[:]
        weights = ensemble.getWeights()
        weights[:, 0, :] = 0.
        ensemble.setWeights(weights)
        ensemble_trimed = trimPDBEnsemble(ensemble, occupancy=0.9)
        assert calcOccupancies(ensemble_trimed).min() > 0.9, \
                     'soft trimPDBEnsemble failed'
        
        ensemble_trimed2 = trimPDBEnsemble(ensemble, occupancy=0.9, hard=True)
        assert calcOccupancies(ensemble_trimed2).min() > 0.9, \
                     'hard trimPDBEnsemble failed'

        atoms = ensemble_trimed.getAtoms()
        atoms2 = ATOMS[ensemble_trimed._indices]
        assert_equal(atoms.getIndices(), atoms2.getIndices(), 
                    'soft trimPDBEnsemble returns a wrong result')

        msa1 = ensemble_trimed.getMSA()
        msa2 = ensemble_trimed2.getMSA()
        assert_equal(msa1.getArray(), msa2.getArray(), 
                    'soft trimPDBEnsemble returns a wrong result')

class TestBuildPDBEnsemble(TestCase):
    
    def testResults(self):
        
        ags = parsePDB([DATA_FILES['3hsy']['path'], 
                        DATA_FILES['3o21']['path'], 
                        DATA_FILES['3p3w']['path']], 
                       subset='ca')
        
        ens1 = buildPDBEnsemble(ags, match_func=bestMatch)
        assert_equal(ens1.numConfs(), 5, 
            'buildPDBEnsemble with bestMatch on full structures did not extract all AMPAR dimers')
        
        ens2 = buildPDBEnsemble(ags, match_func=sameChid)
        assert_equal(ens2.numConfs(), 2, 
            'buildPDBEnsemble with sameChid on full structures did not just take AB dimers')
        
        ens3 = buildPDBEnsemble(ags, match_func=sameChainPos)
        assert_equal(ens3.numConfs(), 2, 
            'buildPDBEnsemble with sameChainPos on full structures did not just take AB dimers')        
   
        
        biomols = parsePDB([DATA_FILES["3hsy"]["path"],
                            DATA_FILES["3o21"]["path"],
                            DATA_FILES["3p3w"]["path"]], 
                        subset="ca",
                        biomol=True, extend_biomol=True)
        
        ens4 = buildPDBEnsemble(biomols, match_func=bestMatch)
        assert_equal(ens4.numConfs(), 5, 
            'buildPDBEnsemble with bestMatch on biomols did not include all AMPAR dimers')    

        ens5 = buildPDBEnsemble(biomols, match_func=sameChid)
        assert_equal(ens5.numConfs(), 2, 
            'buildPDBEnsemble with sameChid on biomols did not just take AB dimer')
        
        ens6 = buildPDBEnsemble(biomols, match_func=bestMatch)
        assert_equal(ens6.numConfs(), 5, 
            'buildPDBEnsemble with sameChainPos on biomols did not include all AMPAR dimers')        
        



"""This module contains unit tests for :mod:`.frame` module."""

from code import interact
from prody.tests import TestCase

from numpy import array
from numpy.testing import assert_allclose

from prody.trajectory import Trajectory
from prody.tests.datafiles import parseDatafile, pathDatafile

DCD = Trajectory(pathDatafile('dcd'))
PDB = parseDatafile('multi_model_truncated', model=1)
RMSD_ALL = array([0.0, 1.380, 1.745])
RMSD_CARBON = array([0.0, 0.964, 1.148])


class TestSuperpose(TestCase):

    def setUp(self):

        DCD.setCoords(PDB.getCoords())
        DCD.reset()

    def testAll(self):

        rmsd = []
        for frame in DCD:
            frame.superpose()
            rmsd.append(frame.getRMSD())
        assert_allclose(rmsd, RMSD_ALL, atol=0.001)

    def testCarbon(self):

        DCD.setAtoms(PDB.carbon)
        rmsd = []
        for frame in DCD:
            frame.superpose()
            rmsd.append(frame.getRMSD())
        assert_allclose(rmsd, RMSD_CARBON, atol=0.001)



"""Align models in a PDB file or multiple structures in separate PDB files."""

from ..apptools import *

__all__ = ['prody_align']

def prody_align(*pdbs, **kwargs):
    """Align models in a PDB file or multiple structures in separate PDB files.
    By default, protein chains will be matched based on selected atoms and
    alignment will be performed based on matching residues.  If non-protein
    atoms are selected and selected atoms match in multiple structures,
    they will be used for alignment.

    :arg pdbs: PDB identifier(s) or filename(s)

    :arg select: atom selection string, default is :term:`calpha`,
        see :ref:`selections`

    :arg model: for NMR files, reference model index, default is ``1``

    :arg seqid: percent sequence identity, default is ``90``

    :arg overlap: percent sequence overlap, default is ``90``

    :arg prefix: prefix for output file, default is PDB filename

    :arg suffix: output filename suffix, default is :file:`_aligned`"""

    from numpy import all
    from prody import LOGGER, writePDB, parsePDB
    from prody import alignCoordsets, printRMSD, matchAlign, superpose

    selstr = kwargs.get('select', 'calpha')
    suffix = kwargs.get('suffix', '_aligned')
    if len(pdbs) == 1:
        pdb = pdbs[0]
        LOGGER.info('Aligning multiple models in: ' + pdb)
        prefix = kwargs.get('prefix')
        model = kwargs.get('model')
        pdb = parsePDB(pdb)
        pdbselect = pdb.select(selstr)
        if pdbselect is None:
            subparser = kwargs.get('subparser')
            if subparser:
                subparser.error('Selection {0} do not match any atoms.'
                                .format(repr(selstr)))
            else:
                raise ValueError('select does not match any atoms')
        LOGGER.info('{0} atoms will be used for alignment.'
                    .format(len(pdbselect)))
        pdbselect.setACSIndex(model-1)
        printRMSD(pdbselect, msg='Before alignment ')
        alignCoordsets(pdbselect)
        printRMSD(pdbselect, msg='After alignment  ')
        outfn = (prefix or pdb.getTitle()) + suffix + '.pdb'
        LOGGER.info('Writing file: ' + outfn)
        writePDB(outfn, pdb)
    else:
        pdbs = list(pdbs)
        reffn = pdbs.pop(0)
        seqid = kwargs.get('seqid')
        overlap = kwargs.get('overlap')
        LOGGER.info('Aligning structures onto: ' + reffn)
        ref = parsePDB(reffn)

        ref_sel = ref.select(selstr)
        if ref_sel:
            LOGGER.info('Selection {0} matched {1} atoms.'
                        .format(repr(selstr), len(ref_sel)))
        else:
            raise ValueError('selection {0} did not match any atoms'
                               .format(repr(selstr)))
        match = True
        if ref_sel.numAtoms('ca') < 2:
            match = False

        for arg in pdbs:
            if arg == reffn:
                continue
            #if '_aligned.pdb' in arg:
            #    continue
            LOGGER.info('Evaluating structure: ' + arg)
            pdb = parsePDB(arg)
            if match:
                result = matchAlign(pdb, ref, seqid=seqid, overlap=overlap,
                                    tarsel=selstr, allcsets=True,
                                    cslabel='Model', csincr=1)
                if result:
                    outfn = pdb.getTitle() + suffix + '.pdb'
                    LOGGER.info('Writing file: ' + outfn)
                    writePDB(outfn, pdb)
                    continue

            pdb_sel = pdb.select(selstr)
            LOGGER.info('Selection {0} matched {1} atoms.'
                        .format(repr(selstr), len(pdb_sel)))
            if (len(pdb_sel) == len(ref_sel) and
                all(pdb_sel.getNames() == ref_sel.getNames())):
                printRMSD(ref_sel, pdb_sel, msg='Before alignment ')
                superpose(pdb_sel, ref_sel)
                printRMSD(ref_sel, pdb_sel, msg='After alignment  ')
                outfn = pdb.getTitle() + suffix + '.pdb'
                LOGGER.info('Writing file: ' + outfn)
                writePDB(outfn, pdb)
            else:
                LOGGER.warn('Failed to align structure ' + arg + '.')

def addCommand(commands):

    subparser = commands.add_parser('align',
        help='align models or structures')

    subparser.add_argument('--quiet', help="suppress info messages to stderr",
        action=Quiet, nargs=0)

    subparser.add_argument('--examples', action=UsageExample, nargs=0,
        help='show usage examples and exit')

    subparser.set_defaults(usage_example=
    """Align models in a PDB structure or multiple PDB structures and save \
aligned coordinate sets.  When multiple structures are aligned, ProDy will \
match chains based on sequence alignment and use best matching pair for aligning the \
structures.

Fetch PDB structure 2k39 and align models (reference model is the first model):

    $ prody align 2k39

Fetch PDB structure 2k39 and align models using backbone of residues with \
number less than 71:

    $ prody align 2k39 --select "backbone and resnum < 71"

Align 1r39 and 1zz2 onto 1p38 using residues with number less than 300:

    $ prody align --select "resnum < 300" 1p38 1r39 1zz2

Align all models of 2k39 onto 1aar using residues 1 to 70 (inclusive):

    $ prody align --select "resnum 1 to 70" 1aar 2k39

Align 1fi7 onto 1hrc using heme atoms:

    $ prody align --select "noh heme and chain A" 1hrc 1fi7""",
    test_examples=[0, 1, 2, 3, 4,]
    )

    group = subparser.add_argument_group('atom/model selection')
    group.add_argument('-s', '--select', dest='select', type=str,
        default='calpha', metavar='SEL',
        help='reference structure atom selection  (default: %(default)s)')
    group.add_argument('-m', '--model', dest='model', type=int,
        default=1, metavar='INT',
        help=('for NMR files, reference model index (default: %(default)s)'))

    group = subparser.add_argument_group('chain matching options')
    group.add_argument('-i', '--seqid', dest='seqid', type=int,
        default=90, metavar='INT',
        help=('percent sequence identity (default: %(default)s)'))
    group.add_argument('-o', '--overlap', dest='overlap', type=int,
        default=90, metavar='INT',
        help=('percent sequence overlap (default: %(default)s)'))

    group = subparser.add_argument_group('output options')
    group.add_argument('-p', '--prefix', dest='prefix', type=str,
        default='', metavar='STR',
        help=('output filename prefix (default: PDB filename)'))
    group.add_argument('-x', '--suffix', dest='suffix', type=str,
        default='_aligned', metavar='STR',
        help=('output filename suffix (default: %(default)s)'))

    subparser.add_argument('pdb', nargs='+',
        help='PDB identifier(s) or filename(s)')

    subparser.set_defaults(func=lambda ns: prody_align(*ns.pdb,
                                                        **ns.__dict__))
    subparser.set_defaults(subparser=subparser)



# -*- coding: utf-8 -*-
"""This module defines :class:`AtomMap` class that allows for pointing atoms in
arbitrary order.

.. _atommaps:

How AtomMap works
===============================================================================

:class:`AtomMap` class adds great flexibility to manipulating atomic data.

First let's see how an instance of :class:`.Selection` (:class:`.Chain`, or
:class:`.Residue`) works.  Below table shows indices for a selection of atoms
in an :class:`~.AtomGroup` and values returned when
:meth:`~.Selection.getNames`, :meth:`~.Selection.getResnames` and
:meth:`~.Selection.getResnums` methods are called.

.. csv-table:: **Atom Subset**
   :header: "Indices", "Names", "Resnames", "Resnums"

   0, N, PHE, 1
   1, CA, PHE, 1
   2, C, PHE, 1
   3, O, PHE, 1
   4, CB, PHE, 1
   5, CG, PHE, 1
   6, CD1, PHE, 1
   7, CD2, PHE, 1
   8, CE1, PHE, 1
   9, CE2, PHE, 1
   10, CZ, PHE, 1

:class:`~.Selection` instances keep indices ordered and do not allow duplicate
values, hence their use is limited. In an :class:`AtomMap`, indices do not need
to be sorted, duplicate indices may exist, even "DUMMY" atoms are allowed.

Let's say we instantiate the following AtomMap::

    amap = AtomMap(atomgroup, indices=[0, 1, 3, 8, 8, 9, 10],
                   mapping=[5, 6, 7, 0, 1, 2, 3])


The size of the :class:`AtomMap` based on this mapping is 8, since the larger
mapping is 7.

Calling the same functions for this AtomMap instance would result in the
following:

.. csv-table:: **Atom Map**
   :header: "Mapping", "Indices", "Names", "Resnames", "Resnums", \
            "MappedFlags", "DummyFlags"

   0, 8, CE1, PHE, 1, 1, 0
   1, 8, CE1, PHE, 1, 1, 0
   2, 9, CE2, PHE, 1, 1, 0
   3, 10, CZ, PHE, 1, 1, 0
   4, , , , 0, 0, 1
   5, 0, N, PHE, 1, 1, 0
   6, 1, CA, PHE, 1, 1, 0
   7, 3, O, PHE, 1, 1, 0

For unmapped atoms, numeric attributes are set to 0, others to empty string,
i.e. ``""``.

.. seealso::
   :class:`AtomMap` are used by :mod:`.proteins` module functions that
   match or map protein chains.  :ref:`pca-xray` and :ref:`pca-dimer`
   examples that make use of these functions and :class:`AtomMap` class.


"""

from numpy import arange, array, ones, zeros, dtype

from prody.utilities import rangeString, copy

from .atom import Atom
from .fields import ATOMIC_FIELDS
from .fields import wrapGetMethod
from .pointer import AtomPointer

__all__ = ['AtomMap']

DUMMY = -1

class AtomMap(AtomPointer):

    """A class for mapping atomic data."""

    __slots__ = ['_ag', '_indices', '_acsi', '_mapping', '_dummies', '_title',
                 '_len', '_idarray']

    def __init__(self, ag, indices, acsi=None, **kwargs):
        """Instantiate an atom map.

        :arg ag: AtomGroup instance from which atoms are mapped
        :arg indices: indices of mapped atoms
        :arg acsi: active coordinate set index, defaults is that of *ag*
        :arg mapping: mapping of atom *indices*
        :arg dummies: dummy atom indices
        :arg title: title of the instance, default is 'Unknown'

        *mapping* and *dummies* arrays must be provided together.  Length of
        *mapping* must be equal to length of *indices*.  Elements of *mapping*
        must be an ordered in ascending order.  When dummy atoms are present,
        number of atoms is the sum of lengths of *mapping* and *dummies*.

        Following built-in functions are customized for this class:

        * :func:`len` returns the number of atoms in the instance.
        * :func:`iter` yields :class:`.Atom` instances.
        * Indexing returns an :class:`.Atom` or an :class:`.AtomMap` instance
          depending on the type and value of the index."""

        AtomPointer.__init__(self, ag, acsi)

        self._dummies = self._mapping = None
        mapping = kwargs.get('mapping', None)
        dummies = kwargs.get('dummies', False)

        try:
            len(dummies)
        except TypeError:
            dummy_array = False
        else:
            dummy_array = True

        if mapping is None:
            if not kwargs.get('intarrays'):
                indices = array(indices, int)
            self._len = len(indices)

            if dummy_array:
                raise TypeError('mapping and dummies must be provided '
                                'together')

            if dummies:
                dummies = (indices == DUMMY).nonzero()[0]
                if len(dummies):
                    self._dummies = dummies
                    self._mapping = (indices > DUMMY).nonzero()[0]
                    self._indices = indices[self._mapping]
                    self._idarray = indices
                else:
                    self._indices = self._idarray = indices
            else:
                self._indices = self._idarray = indices
        else:
            if dummies is None:
                raise TypeError('mapping and dummies must be provided '
                                'together')

            if len(mapping) != len(indices):
                raise ValueError('indices and mapping arrays must have the '
                                 'same length')
            if not kwargs.get('intarrays'):
                indices = array(indices, int)
                mapping = array(mapping, int)
                if dummy_array:
                    dummies = array(dummies, int)
            #if any(mapping[1:] - mapping[:-1] < 0):
            #    raise ValueError('mapping must be an ordered array')
            self._len = len(indices)
            if dummy_array:
                self._indices = indices
                self._mapping = mapping
                self._dummies = dummies
                self._len += len(dummies)
                self._idarray = idarray = zeros(self._len, int)
                idarray[self._mapping] = self._indices
                idarray[self._dummies] = DUMMY
            else:
                self._indices = self._idarray = indices[mapping]
        self._title = str(kwargs.get('title', 'Unknown'))

    def __repr__(self):

        rep = '<AtomMap: {0} from {1} ({2} atoms'.format(
                self._title, self._ag.getTitle(), self._len)
        if self.numDummies():
            rep += ', {0} mapped, {1} dummy'.format(self.numMapped(),
                                                        self.numDummies())

        n_csets = self._ag.numCoordsets()
        if n_csets > 1:
            rep += '; active #{0} of {1} coordsets)>'.format(
                    self.getACSIndex(), n_csets)
        elif n_csets == 0:
            rep += '; no coordinates'
        return rep + ')>'

    def __str__(self):

        return 'AtomMap {0}'.format(self._title)

    def __len__(self):

        return self._len

    def __getitem__(self, index):

        indices = self._idarray[index]
        try:
            len(indices)
        except TypeError:
            if indices != DUMMY:
                return self._ag[indices]
        else:
            return AtomMap(self._ag, indices, self._acsi,
                       title='({0})[{1}]'.format(self._title, repr(index)),
                       intarrays=True, dummies=self.numDummies())

    def getTitle(self):
        """Returns title of the instance."""

        return self._title

    def setTitle(self, title):
        """Set title of the instance."""

        self._title = str(title)

    def numAtoms(self, flag=None):
        """Returns number of atoms."""

        return len(self._getSubset(flag)) if flag else self._len

    def iterAtoms(self):
        """Yield atoms, and **None** for dummies."""

        ag = self._ag
        acsi = self.getACSIndex()
        for index in self.getIndices():
            yield Atom(ag, index, acsi) if index > DUMMY else None

    __iter__ = iterAtoms

    def getCoords(self):
        """Returns a copy of coordinates from the active coordinate set."""

        coords = self._ag._getCoordsets()
        if coords is not None:
            if self._mapping is None:
                xyz = coords[self.getACSIndex(), self._indices]
            else:
                xyz = zeros((self._len, 3), float)
                xyz[self._mapping] = coords[self.getACSIndex(), self._indices]
            return xyz

    _getCoords = getCoords

    def setCoords(self, coords):
        """Set coordinates of atoms in the active coordinate set."""

        coordsets = self._ag._getCoordsets()
        if coordsets is not None:
            if self._mapping is None:
                coordsets[self.getACSIndex(), self._indices] = coords
            elif self._dummies is None:
                coordsets[self.getACSIndex(), self._indices] = coords


    def getCoordsets(self, indices=None):
        """Returns coordinate set(s) at given *indices*, which may be an integer
        or a list/array of integers."""

        coords = self._ag._getCoordsets()
        if coords is not None:
            if indices is None:
                coords = coords[:, self._indices]
            else:
                try:
                    len(indices)
                except TypeError:
                    coords = coords[indices, self._indices]
                else:
                    coords = coords[indices][:, self._indices]

            if self._mapping is None:
                return coords
            else:
                csets = zeros(coords.shape[:-2] + (self._len, 3))
                csets[:, self._mapping] = coords
                return csets

    _getCoordsets = getCoordsets

    def iterCoordsets(self):
        """Yield copies of coordinate sets."""

        coords = self._ag._getCoordsets()
        if coords is not None:
            mapping = self._mapping
            n_atoms = self._len
            indices = self._indices
            for i in range(self._ag.numCoordsets()):
                xyz = zeros((n_atoms, 3), float)
                xyz[mapping] = coords[i, indices]
                yield xyz

    _iterCoordsets = iterCoordsets

    def getData(self, label):
        """Returns a copy of data associated with *label*, if it is present."""

        data = self._ag._getData(label)
        if data is not None:
            result = zeros((self._len,) + data.shape[1:], data.dtype)
            result[self._mapping] = data[self._indices]
            return result

    _getData = getData

    def getFlags(self, label):
        """Returns a copy of atom flags for given *label*, or **None** when
        flags for *label* is not set."""

        if label == 'dummy':
            flags = zeros(self._len, bool)
            if self._dummies is not None:
                flags[self._dummies] = True
        elif label == 'mapped':
            flags = ones(self._len, bool)
            if self._dummies is not None:
                flags[self._dummies] = False
        else:
            flags = None
            agflags = self._ag._getFlags(label)
            if agflags is not None:
                flags = zeros(self._len, bool)
                flags[self._mapping] = agflags[self._indices]
        return flags

    _getFlags = getFlags

    def _getSubset(self, label):

        return self._idarray[self._getFlags(label)]

    def getIndices(self):
        """Returns a copy of indices of atoms, with maximum integer value
        dummies."""

        return copy(self._idarray)

    def _getIndices(self):
        """Returns indices of atoms, with maximum integer value dummies."""

        return self._idarray

    def getMapping(self):
        """Returns a copy of mapping of indices."""

        mapping = self._mapping
        return arange(self._len) if mapping is None else mapping.copy()

    def _getMapping(self):
        """Returns mapping of indices."""

        mapping = self._mapping
        return arange(self._len) if mapping is None else mapping

    def numMapped(self):
        """Returns number of mapped atoms."""

        return len(self._indices)

    def numDummies(self):
        """Returns number of dummy atoms."""

        return 0 if self._dummies is None else len(self._dummies)

    def getSelstr(self):
        """Returns selection string that selects mapped atoms."""

        return 'index ' + rangeString(self._indices)

    def getHierView(self, **kwargs):
        """Returns a hierarchical view of the this chain."""

        return HierView(self, **kwargs)


for fname, field in ATOMIC_FIELDS.items():

    if field.private:
        continue

    meth = field.meth_pl
    getMeth = 'get' + meth

    def getData(self, meth=field.meth_pl, dtype=field.dtype):
        data = getattr(self._ag, '_get' + meth)()
        if data is not None:
            if self._mapping is None:
                return data[self._indices]
            else:
                result = zeros((self._len,) + data.shape[1:], dtype)
                result[self._mapping] = data[self._indices]
                return result

    getData = wrapGetMethod(getData)
    getData.__name__ = getMeth
    getData.__doc__ = (field.getDocstr('get', selex=False) +
                       ' Entries for dummy atoms will be ``{0}``.'
                       .format(repr(dtype(field.dtype).type())))
    setattr(AtomMap, getMeth, getData)
    setattr(AtomMap, '_' + getMeth, getData)

del getData



# -*- coding: utf-8 -*-
"""This module defines a class and a function for anisotropic network model
(ANM) calculations."""

import numpy as np
from numbers import Integral

from prody import LOGGER
from prody.atomic import Atomic, AtomGroup
from prody.proteins import parsePDB
from prody.utilities import checkCoords, solveEig
from prody.kdtree import KDTree

from .nma import NMA, MaskedNMA
from .gnm import GNMBase, checkENMParameters

__all__ = ['ANM', 'MaskedANM', 'calcANM']

class ANMBase(NMA):

    def __init__(self, name='Unknown'):

        super(ANMBase, self).__init__(name)
        self._is3d = True
        self._cutoff = None
        self._gamma = None
        self._hessian = None

    def _reset(self):

        super(ANMBase, self)._reset()
        self._cutoff = None
        self._gamma = None
        self._hessian = None
        self._is3d = True
    
    def _clear(self):
        self._trace = None
        self._cov = None

    def getHessian(self):
        """Returns a copy of the Hessian matrix."""

        if self._hessian is None:
            return None
        return self._getHessian().copy()

    def _getHessian(self):
        """Returns the Hessian matrix."""

        return self._hessian

    def setHessian(self, hessian):
        """Set Hessian matrix.  A symmetric matrix is expected, i.e. not a
        lower- or upper-triangular matrix."""

        if not isinstance(hessian, np.ndarray):
            raise TypeError('hessian must be a Numpy array')
        elif hessian.ndim != 2 or hessian.shape[0] != hessian.shape[1]:
            raise ValueError('hessian must be square matrix')
        elif hessian.shape[0] % 3:
            raise ValueError('hessian.shape must be (3*n_atoms,3*n_atoms)')
        elif hessian.dtype != float:
            try:
                hessian = hessian.astype(float)
            except:
                raise ValueError('hessian.dtype must be float')
        self._reset()
        self._hessian = hessian
        self._dof = hessian.shape[0]
        self._n_atoms = self._dof // 3
        
    def buildHessian(self, coords, cutoff=15., gamma=1., **kwargs):
        """Build Hessian matrix for given coordinate set.

        :arg coords: a coordinate set or an object with ``getCoords`` method
        :type coords: :class:`numpy.ndarray`, :class:`Atomic`, 
            :class:`Ensemble`, :class:`Trajectory`

        :arg cutoff: cutoff distance (Å) for pairwise interactions,
            default is 15.0 Å, minimum is 4.0 Å
        :type cutoff: float

        :arg gamma: spring constant, default is 1.0
        :type gamma: float, :class:`Gamma`

        :arg sparse: elect to use sparse matrices, default is **False**. If
            Scipy is not found, :class:`ImportError` is raised.
        :type sparse: bool

        :arg kdtree: elect to use KDTree for building Hessian matrix,
            default is **False** since KDTree method is slower
        :type kdtree: bool

        Instances of :class:`Gamma` classes and custom functions are
        accepted as *gamma* argument.

        When Scipy is available, user can select to use sparse matrices for
        efficient usage of memory at the cost of computation speed.
        
        Any atoms or points can be used for building a Hessian matrix, including 
        calphas, phosphorus and carbon atoms from nucleic acids, all atoms, or 
        pseudoatoms fitted to density maps with algorithms such as TRN. 
        
        The cutoff distance may need to be adjusted depending on the coarse graining 
        level of the atoms or points used."""

        try:
            coords = (coords._getCoords() if hasattr(coords, '_getCoords') else
                      coords.getCoords())
        except AttributeError:
            try:
                checkCoords(coords)
            except TypeError:
                raise TypeError('coords must be a Numpy array or an object '
                                'with `getCoords` method')

        cutoff, g, gamma = checkENMParameters(cutoff, gamma)
        self._reset()
        self._cutoff = cutoff
        self._gamma = g
        n_atoms = coords.shape[0]

        dof = n_atoms * 3
        LOGGER.timeit('_anm_hessian')

        sparse = kwargs.get('sparse', False)
        if sparse:
            try:
                from scipy import sparse as scipy_sparse
            except ImportError:
                raise ImportError('failed to import scipy.sparse, which  is '
                                  'required for sparse matrix calculations')
            kirchhoff = scipy_sparse.lil_matrix((n_atoms, n_atoms))
            hessian = scipy_sparse.lil_matrix((dof, dof))
        else:
            kirchhoff = np.zeros((n_atoms, n_atoms), 'd')
            hessian = np.zeros((dof, dof), float)

        if kwargs.get('kdtree', False):
            LOGGER.info('Using KDTree for building the Hessian.')
            kdtree = KDTree(coords)
            kdtree.search(cutoff)
            for i, j in kdtree.getIndices():
                i2j = coords[j] - coords[i]
                dist2 = np.dot(i2j, i2j)
                g = gamma(dist2, i, j)
                super_element = np.outer(i2j, i2j) * (- g / dist2)
                res_i3 = i*3
                res_i33 = res_i3+3
                res_j3 = j*3
                res_j33 = res_j3+3
                hessian[res_i3:res_i33, res_j3:res_j33] = super_element
                hessian[res_j3:res_j33, res_i3:res_i33] = super_element
                hessian[res_i3:res_i33, res_i3:res_i33] = \
                    hessian[res_i3:res_i33, res_i3:res_i33] - super_element
                hessian[res_j3:res_j33, res_j3:res_j33] = \
                    hessian[res_j3:res_j33, res_j3:res_j33] - super_element
                kirchhoff[i, j] = -g
                kirchhoff[j, i] = -g
                kirchhoff[i, i] = kirchhoff[i, i] + g
                kirchhoff[j, j] = kirchhoff[j, j] + g
        else:
            cutoff2 = cutoff * cutoff
            for i in range(n_atoms):
                res_i3 = i*3
                res_i33 = res_i3+3
                i_p1 = i+1
                i2j_all = coords[i_p1:, :] - coords[i]
                for j, dist2 in enumerate((i2j_all ** 2).sum(1)):
                    if dist2 > cutoff2:
                        continue
                    i2j = i2j_all[j]
                    j += i_p1
                    g = gamma(dist2, i, j)
                    res_j3 = j*3
                    res_j33 = res_j3+3
                    super_element = np.outer(i2j, i2j) * (- g / dist2)
                    hessian[res_i3:res_i33, res_j3:res_j33] = super_element
                    hessian[res_j3:res_j33, res_i3:res_i33] = super_element
                    hessian[res_i3:res_i33, res_i3:res_i33] = \
                        hessian[res_i3:res_i33, res_i3:res_i33] - super_element
                    hessian[res_j3:res_j33, res_j3:res_j33] = \
                        hessian[res_j3:res_j33, res_j3:res_j33] - super_element
                    kirchhoff[i, j] = -g
                    kirchhoff[j, i] = -g
                    kirchhoff[i, i] = kirchhoff[i, i] + g
                    kirchhoff[j, j] = kirchhoff[j, j] + g

        if sparse:
            kirchhoff = kirchhoff.tocsr()
            hessian = hessian.tocsr()

        LOGGER.report('Hessian was built in %.2fs.', label='_anm_hessian')
        self._kirchhoff = kirchhoff
        self._hessian = hessian
        self._n_atoms = n_atoms
        self._dof = dof

    def calcModes(self, n_modes=20, zeros=False, turbo=True, **kwargs):
        """Calculate normal modes.  This method uses :func:`scipy.linalg.eigh`
        function to diagonalize the Hessian matrix. When Scipy is not found,
        :func:`numpy.linalg.eigh` is used.

        :arg n_modes: number of non-zero eigenvalues/vectors to calculate.
            If **None** or ``'all'`` is given, all modes will be calculated.
        :type n_modes: int or None, default is 20

        :arg zeros: If **True**, modes with zero eigenvalues will be kept.
        :type zeros: bool, default is **True**

        :arg turbo: Use a memory intensive, but faster way to calculate modes.
        :type turbo: bool, default is **True**

        :arg nproc: number of processors for thread pool limit,
            default is **0**, meaning don't impose limit
        :type nproc: int
        """

        if self._hessian is None:
            raise ValueError('Hessian matrix is not built or set')
        if str(n_modes).lower() == 'all':
            n_modes = None
        assert n_modes is None or isinstance(n_modes, int) and n_modes > 0, \
            'n_modes must be a positive integer'
        assert isinstance(zeros, bool), 'zeros must be a boolean'
        assert isinstance(turbo, bool), 'turbo must be a boolean'
        self._clear()
        LOGGER.timeit('_anm_calc_modes')
        values, vectors, vars = solveEig(self._hessian, n_modes=n_modes, zeros=zeros, 
                                         turbo=turbo, expct_n_zeros=6, **kwargs)
        self._eigvals = values
        self._array = vectors
        self._vars = vars
        self._trace = self._vars.sum()

        self._n_modes = len(self._eigvals)
        if self._n_modes > 1:
            LOGGER.report('{0} modes were calculated in %.2fs.'
                        .format(self._n_modes), label='_anm_calc_modes')
        else:
            LOGGER.report('{0} mode was calculated in %.2fs.'
                        .format(self._n_modes), label='_anm_calc_modes')

    
    def setEigens(self, vectors, values=None):
        self._clear()
        super(ANMBase, self).setEigens(vectors, values)
        

class ANM(ANMBase, GNMBase):

    """Class for Anisotropic Network Model (ANM) analysis of proteins
    ([PD00]_, [ARA01]_).

    See a usage example in :ref:`anm`.

    .. [PD00] Doruker P, Atilgan AR, Bahar I. Dynamics of proteins predicted by
       molecular dynamics simulations and analytical approaches: Application to
       a-amylase inhibitor. *Proteins* **2000** 40:512-524.

    .. [ARA01] Atilgan AR, Durrell SR, Jernigan RL, Demirel MC, Keskin O,
       Bahar I. Anisotropy of fluctuation dynamics of proteins with an
       elastic network model. *Biophys. J.* **2001** 80:505-515."""


    def __init__(self, name='Unknown'):

        super(ANM, self).__init__(name)

class MaskedANM(ANM, MaskedNMA):
    def __init__(self, name='Unknown', mask=False, masked=True):
        ANM.__init__(self, name)
        MaskedNMA.__init__(self, name, mask, masked)

    def calcModes(self, n_modes=20, zeros=False, turbo=True):
        self._maskedarray = None
        super(MaskedANM, self).calcModes(n_modes, zeros, turbo)

    def _reset(self):
        super(MaskedANM, self)._reset()
        self._maskedarray = None

    def setHessian(self, hessian):
        """Set Hessian matrix.  A symmetric matrix is expected, i.e. not a
        lower- or upper-triangular matrix."""

        if not isinstance(hessian, np.ndarray):
            raise TypeError('hessian must be a Numpy array')
        elif hessian.ndim != 2 or hessian.shape[0] != hessian.shape[1]:
            raise ValueError('hessian must be square matrix')
        elif hessian.shape[0] % 3:
            raise ValueError('hessian.shape must be (3*n_atoms,3*n_atoms)')
        elif hessian.dtype != float:
            try:
                hessian = hessian.astype(float)
            except:
                raise ValueError('hessian.dtype must be float')
        
        mask = self.mask
        if not self.masked:
            if not np.isscalar(mask):
                if self.is3d():
                    mask = np.repeat(mask, 3)
                try:
                    hessian = hessian[mask, :][:, mask]
                except IndexError:
                    raise IndexError('size mismatch between Hessian (%d) and mask (%d).'
                                     'Try set masked to False or reset mask'%(len(hessian), len(mask)))

        super(MaskedANM, self).setHessian(hessian)

    def _getHessian(self):
        """Returns the Hessian matrix."""

        hessian = self._hessian

        if hessian is None: return None

        if not self._isOriginal():
            hessian = self._extend(hessian, axis=None)

        return hessian


def calcANM(pdb, selstr='calpha', cutoff=15., gamma=1., n_modes=20,
            zeros=False, title=None):
    """Returns an :class:`ANM` instance and atoms used for the calculations.
    By default only alpha carbons are considered, but selection string helps
    selecting a subset of it.  *pdb* can be a PDB code, :class:`.Atomic` 
    instance, or a Hessian matrix (:class:`~numpy.ndarray`)."""

    if isinstance(pdb, np.ndarray):
        H = pdb
        if title is None:
            title = 'Unknown'
        anm = ANM(title)
        anm.setHessian(H)
        anm.calcModes(n_modes, zeros)
        
        return anm
        
    else:
        if isinstance(pdb, str):
            ag = parsePDB(pdb)
            if title is None:
                title = ag.getTitle()
        elif isinstance(pdb, Atomic):
            ag = pdb
            if title is None:
                if isinstance(pdb, AtomGroup):
                    title = ag.getTitle()
                else:
                    title = ag.getAtomGroup().getTitle()
        else:
            raise TypeError('pdb must be an atomic class, not {0}'
                            .format(type(pdb)))
        
        anm = ANM(title)
        sel = ag.select(selstr)
        anm.buildHessian(sel, cutoff, gamma)
        anm.calcModes(n_modes, zeros)
    
        return anm, sel



# -*- coding: utf-8 -*-
"""This module defines functions for Mechanical Stiffness calculations."""

from numbers import Integral

import numpy as np

from prody import LOGGER
from prody.utilities import checkCoords


__all__ = ['calcMechStiff', 'calcStiffnessRange', 'calcMechStiffStatistic', 
           'calcStiffnessRangeSel']

def calcMechStiff(modes, coords, kbt=1.):
    """Calculate stiffness matrix calculated using :class:`.ANM` instance. 
    Method described in [EB08]_ and [KMR17]_. 

    .. [KMR] Mikulska-Ruminska K., Kulik A.J., Benadiba C., Bahar I., Dietler G., Nowak W.
        Nanomechanics of multidomain neuronal cell adhesion protein contactin revealed by 
        single molecule AFM and SMD. *Sci Rep* **2017** 7:8852. 

    :arg coords: a coordinate set or an object with ``getCoords`` method
    :type coords: :class:`numpy.ndarray`.

    :arg n_modes: number of non-zero eigenvalues/vectors to calculate.
        If **None** is given, all modes will be calculated (3x number of atoms).
    :type n_modes: int or **None**, default is 20.
    
    Authors: Mustafa Tekpinar & Karolina Mikulska-Ruminska & Cihan Kaya
    """

    try:
        coords = (coords._getCoords() if hasattr(coords, '_getCoords') else
                    coords.getCoords())
    except AttributeError:
        try:
            checkCoords(coords)
        except TypeError:
            raise TypeError('coords must be a Numpy array or an object '
                            'with `getCoords` method')
    try:
        is3d = modes.is3d()
        eigvecs = modes.getArray().T.flatten()
        eigvals = modes.getEigvals()
    except:
        raise TypeError('modes must be either an NMA or ModeSet object')

    if not is3d:
        raise TypeError('modes must be 3-dimensional')

    n_atoms = modes.numAtoms()
    n_modes = modes.numModes()
    
    LOGGER.timeit('_sm')

    sm = np.zeros((n_atoms, n_atoms), np.double)
    from .smtools import calcSM
    LOGGER.info('Calculating stiffness matrix.')

    calcSM(coords, sm, eigvecs, eigvals,
            n_atoms, n_modes, float(kbt))

    LOGGER.report('Stiffness matrix calculated in %.2lfs.', label='_sm')
    
    LOGGER.info('The range of effective force constant is: {0} to {1}.'
                                .format(*calcStiffnessRange(sm)))

    return sm


def calcStiffnessRange(stiffness):
    """ Return the range of effective spring constant."""
    
    return np.min(stiffness[np.nonzero(stiffness)]), np.amax(stiffness)

def calcMechStiffStatistic(stiffness, rangeK, minAA=0, AA='all'):
    """Returns number of effective spring constant with set range of
    amino acids of protein structure.
    ``AA`` can be a list with a range of analysed amino acids as:
    [first_aa, last_aa, first_aa2, last_aa2],
    minAA - eliminate amino acids that are within 20aa and
    ``rangeK`` is a list [minK, maxK]"""
    
    if AA == 'all':
        sm = stiffness
    elif isinstance(AA, Integral): 
        sm = stiffness[0: AA, (-1)*AA-1:-1]
    elif not np.isscalar(AA) and len(AA) == 1:
        sm = stiffness[0: AA, (-1)*AA-1:-1]
    elif not np.isscalar(AA) and len(AA) == 4:
        sm = stiffness[AA[0]:AA[1],AA[2]:AA[3]]
    if minAA > 0:
        sm2 = sm[minAA:-1,0:-1-minAA]  # matrix without close contacts
        sm3 = np.tril(sm2, k=-1)
        #sort_sm2 = np.sort((np.tril(sm2, k=-1)1).flatten())
        a = np.where(np.logical_and(sm3>rangeK[0], sm3<rangeK[1]))
    if minAA == 0:
        sm2 = np.tril(sm, k=-1)
        a = np.where(np.logical_and(sm2>rangeK[0], sm2<rangeK[1]))
    return len(a[0])
    

def calcStiffnessRangeSel(stiffness, value, minAA=20, AA='all'):
    """Returns minimum or maximum value of sping constant from 
    mechanical stiffness calculations for residues that are within 
    more than ``min_aa`` from each other. ``Value`` should be 'minK'
    or 'maxK'. It alow to avoid residues near each other. 
    ``AA`` is a number of residues from both terminus (N and C)
    of protein strcuture, it can be ``all`` or int value (than first 
    and last ``AA`` residues will be analysed. 
    With ``minAA=0`` it can be used to search the highest/lowest 
    values of interactions between N-C terminus if protein structure
    has a shear, zipper or SD1-disconnected mechanical clamp 
    -it is common in FnIII/Ig like domains and determines the maximum 
    unfolding force in AFM or SMD method."""
    
    if AA == 'all':
        sm = stiffness
    elif isinstance(AA, Integral):
        sm = stiffness[0: AA, (-1)*AA-1:-1]
    minK = np.min(sm[np.nonzero(sm)]) 
    maxK = np.amax(sm)
    
    if value == 'minK':
        indices = np.where(sm == minK)
    elif value == 'maxK':
        indices = np.where(sm == maxK)
    try:
        residue_diff = abs(indices[0][0]-indices[0][1])
    except: 
        residue_diff = abs(indices[0]-indices[1])
    sm_mod = sm.copy()
    checking = True
    while checking:
        if residue_diff < minAA:
            sm_mod[indices[0][0],indices[0][1]] = 0
            sm_mod[indices[1][0],indices[1][1]] = 0
            if value == 'minK':
                mK = np.min(sm_mod[np.nonzero(sm_mod)])
                indices = np.where(sm_mod == np.min(sm_mod[np.nonzero(sm_mod)]))
            elif value == 'maxK':
                mK = np.amax(sm_mod)
                indices = np.where(sm_mod == np.amax(sm_mod))
            try:
                residue_diff = abs(indices[0][0]-indices[0][1])
            except: residue_diff = abs(indices[0]-indices[1])
        else: 
            if value == 'minK':
                mK = np.min(sm_mod[np.nonzero(sm_mod)])
                indices = np.where(sm_mod == np.min(sm_mod[np.nonzero(sm_mod)]))
            elif value == 'maxK':
                mK = np.amax(sm_mod)
                indices = np.where(sm_mod == np.amax(sm_mod))
            checking = False

    if len(indices[0]) == 2:
        return mK, list(indices[0])
    return mK, list(indices[0])+list(indices[1])
    


# -*- coding: utf-8 -*-
"""This module defines a base class for format specific trajectory classes."""

from os.path import isfile, abspath, split, splitext
from numbers import Integral

import numpy as np

from prody import LOGGER
from prody.utilities import relpath

from .trajbase import TrajBase

__all__ = ['TrajFile']


class TrajFile(TrajBase):

    """A base class for trajectory file classes:

      * :class:`.DCDFile`"""


    def __init__(self, filename, mode='r'):
        """Open *filename* for reading (default, ``mode="r"``), writing
        (``mode="w"``), or appending (``mode="r+"`` or ``mode="a"``).
        Binary mode option will be appended automatically."""

        if not isinstance(filename, str):
            raise TypeError("filename argument must be a string")
        if not isinstance(mode, str):
            TypeError('mode argument must be string')
        if not mode in ('r', 'w', 'a', 'r+'):
            ValueError("mode string must begin with one of 'r', 'w', 'r+', or "
                       "'a'")
        name = splitext(split(filename)[1])[0]
        TrajBase.__init__(self, name)
        self._file = None
        if mode == 'r' and not isfile(filename):
            raise IOError("[Errno 2] No such file or directory: '{0}'"
                          .format(filename))
        self._filename = filename
        if mode in ('a', 'r+'):
            self._file = open(filename, 'r+b')
            self._file.seek(0)
            mode = 'a'
        else:
            if not mode.endswith('b'):
                mode += 'b'
            self._file = open(filename, mode)

        self._mode = mode
        self._bytes_per_frame = None
        self._first_byte = None
        self._dtype = np.float32

        self._timestep = 1
        self._first_ts = 0
        self._framefreq = 1
        self._n_fixed = 0

    def __del__(self):

        if self._file is not None:
            self._file.close()

    def __repr__(self):

        if self._closed:
            return ('<{0}: {1} (closed)>').format(
                        self.__class__.__name__, self._title)

        link = ''
        if self._ag is not None:
            link = 'linked to ' + str(self._ag) + '; '

        if self._mode.startswith('r'):
            next = 'next {0} of {1} frames; '.format(self._nfi,
                                                       self._n_csets)
        else:
            next = '{0} frames written; '.format(self._n_csets)

        if self._indices is None:
            atoms = '{0} atoms'.format(self._n_atoms)
        else:
            atoms = 'selected {0} of {1} atoms'.format(
                                        self.numSelected(), self._n_atoms)

        return '<{0}: {1} ({2}{3}{4})>'.format(
                   self.__class__.__name__, self._title, link, next, atoms)

    def getFilename(self, absolute=False):
        """Returns relative path to the current file. For absolute path,
        pass ``absolute=True`` argument."""

        if absolute:
            return abspath(self._filename)
        return relpath(self._filename)

    def getFrame(self, index):
        """Returns frame at given *index*."""

        if self._closed:
            raise ValueError('I/O operation on closed file')
        if not isinstance(index, Integral):
            raise IndexError('index must be an integer')
        if not 0 <= index < self._n_csets:
            raise IndexError('index must be greater or equal to 0 and less '
                             'than number of frames')
        nfi = self._nfi
        if index > nfi:
            self.skip(index - nfi)
        elif index < nfi:
            self.reset()
            if index > 0:
                self.skip(index)
        return next(self)

    getFrame.__doc__ = TrajBase.getFrame.__doc__

    def getCoordsets(self, indices=None):

        if self._closed:
            raise ValueError('I/O operation on closed file')
        if indices is None:
            indices = np.arange(self._n_csets)
        elif isinstance(indices, int):
            indices = np.array([indices])
        elif isinstance(indices, slice):
            indices = np.arange(*indices.indices(self._n_csets))
            indices.sort()
        elif isinstance(indices, (list, np.ndarray)):
            indices = np.unique(indices)
        else:
            raise TypeError('indices must be an integer or a list of integers')

        nfi = self._nfi
        self.reset()

        n_atoms = self.numSelected()
        coords = np.zeros((len(indices), n_atoms, 3), self._dtype)

        prev = 0
        next = self.nextCoordset
        for i, index in enumerate(indices):
            diff = index - prev
            if diff > 1:
                self.skip(diff-1)
            xyz = next()
            if xyz is None:
                LOGGER.warning('Expected {0} frames, but parsed {1}.'
                               .format(len(indices), i))
                self.goto(nfi)
                return coords[:i]
            coords[i] = xyz
            prev = index

        self.goto(nfi)
        return coords

    getCoordsets.__doc__ = TrajBase.getCoordsets.__doc__

    def skip(self, n):

        if self._closed:
            raise ValueError('I/O operation on closed file')
        if not isinstance(n, Integral):
            raise ValueError('n must be an integer')
        if n > 0:
            left = self._n_csets - self._nfi
            if n > left:
                n = left
            self._file.seek(n * self._bytes_per_frame, 1)
            self._nfi += n

    skip.__doc__ = TrajBase.skip.__doc__

    def goto(self, n):

        if self._closed:
            raise ValueError('I/O operation on closed file')
        if not isinstance(n, int):
            raise ValueError('n must be an integer')
        n_csets = self._n_csets
        if n == 0:
            self.reset()
        else:
            if n < 0:
                n = n_csets + n
            if n < 0:
                n = 0
            elif n > n_csets:
                n = n_csets
            self._file.seek(self._first_byte + n * self._bytes_per_frame)
            self._nfi = n

    goto.__doc__ = TrajBase.goto.__doc__

    def reset(self):

        if self._closed:
            raise ValueError('I/O operation on closed file')
        self._file.seek(self._first_byte)
        self._nfi = 0

    reset.__doc__ = TrajBase.reset.__doc__

    def close(self):

        self._file.close()
        self._nfi = 0
        self._closed = True

    close.__doc__ = TrajBase.close.__doc__

    def getTimestep(self):
        """Returns timestep size."""

        return self._timestep

    def getFirstTimestep(self):
        """Returns first timestep value."""

        return self._first_ts

    def getFrameFreq(self):
        """Returns timesteps between frames."""

        return self._framefreq

    def numFixed(self):
        """Returns number of fixed atoms."""

        return self._n_fixed



# Copyright (C) 2013 by Yanbo Ye (yeyanbo289@gmail.com)
#
# This file is originally part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
# 
# The file is adapted by She Zhang (shezhang620@gmail.com) on 04/13/2020 to fix a 
# bug in the construction of the UPGMA tree and made suitable for being used by 
# ProDy.

"""Classes and methods for tree construction."""

import itertools
import copy
import numbers
from Bio.Phylo import BaseTree
from Bio.Phylo.TreeConstruction import _DistanceMatrix as DistanceMatrix

__all__ = ['_Matrix', 'DistanceMatrix', 'TreeConstructor', 'DistanceTreeConstructor']

class _Matrix:
    """Base class for distance matrix or scoring matrix.

    Accepts a list of names and a lower triangular matrix.::

        matrix = [[0],
                  [1, 0],
                  [2, 3, 0],
                  [4, 5, 6, 0]]
        represents the symmetric matrix of
        [0,1,2,4]
        [1,0,3,5]
        [2,3,0,6]
        [4,5,6,0]

    :Parameters:
        names : list
            names of elements, used for indexing
        matrix : list
            nested list of numerical lists in lower triangular format

    Examples
    --------
    >>> from Bio.Phylo.TreeConstruction import _Matrix
    >>> names = ['Alpha', 'Beta', 'Gamma', 'Delta']
    >>> matrix = [[0], [1, 0], [2, 3, 0], [4, 5, 6, 0]]
    >>> m = _Matrix(names, matrix)
    >>> m
    _Matrix(names=['Alpha', 'Beta', 'Gamma', 'Delta'], matrix=[[0], [1, 0], [2, 3, 0], [4, 5, 6, 0]])

    You can use two indices to get or assign an element in the matrix.

    >>> m[1,2]
    3
    >>> m['Beta','Gamma']
    3
    >>> m['Beta','Gamma'] = 4
    >>> m['Beta','Gamma']
    4

    Further more, you can use one index to get or assign a list of elements related to that index.

    >>> m[0]
    [0, 1, 2, 4]
    >>> m['Alpha']
    [0, 1, 2, 4]
    >>> m['Alpha'] = [0, 7, 8, 9]
    >>> m[0]
    [0, 7, 8, 9]
    >>> m[0,1]
    7

    Also you can delete or insert a column&row of elemets by index.

    >>> m
    _Matrix(names=['Alpha', 'Beta', 'Gamma', 'Delta'], matrix=[[0], [7, 0], [8, 4, 0], [9, 5, 6, 0]])
    >>> del m['Alpha']
    >>> m
    _Matrix(names=['Beta', 'Gamma', 'Delta'], matrix=[[0], [4, 0], [5, 6, 0]])
    >>> m.insert('Alpha', [0, 7, 8, 9] , 0)
    >>> m
    _Matrix(names=['Alpha', 'Beta', 'Gamma', 'Delta'], matrix=[[0], [7, 0], [8, 4, 0], [9, 5, 6, 0]])

    """

    def __init__(self, names, matrix=None):
        """Initialize matrix.

        Arguments are a list of names, and optionally a list of lower
        triangular matrix data (zero matrix used by default).
        """
        # check names
        if isinstance(names, list) and all(isinstance(s, str) for s in names):
            if len(set(names)) == len(names):
                self.names = names
            else:
                raise ValueError("Duplicate names found")
        else:
            raise TypeError("'names' should be a list of strings")

        # check matrix
        if matrix is None:
            # create a new one with 0 if matrix is not assigned
            matrix = [[0] * i for i in range(1, len(self) + 1)]
            self.matrix = matrix
        else:
            # check if all elements are numbers
            if (
                isinstance(matrix, list)
                and all(isinstance(l, list) for l in matrix)
                and all(
                    isinstance(n, numbers.Number)
                    for n in [item for sublist in matrix for item in sublist]
                )
            ):
                # check if the same length with names
                if len(matrix) == len(names):
                    # check if is lower triangle format
                    if [len(m) for m in matrix] == list(range(1, len(self) + 1)):
                        self.matrix = matrix
                    else:
                        raise ValueError("'matrix' should be in lower triangle format")
                else:
                    raise ValueError("'names' and 'matrix' should be the same size")
            else:
                raise TypeError("'matrix' should be a list of numerical lists")

    def __getitem__(self, item):
        """Access value(s) by the index(s) or name(s).

        For a _Matrix object 'dm'::

            dm[i]                   get a value list from the given 'i' to others;
            dm[i, j]                get the value between 'i' and 'j';
            dm['name']              map name to index first
            dm['name1', 'name2']    map name to index first

        """
        # Handle single indexing
        if isinstance(item, (int, str)):
            index = None
            if isinstance(item, int):
                index = item
            elif isinstance(item, str):
                if item in self.names:
                    index = self.names.index(item)
                else:
                    raise ValueError("Item not found.")
            else:
                raise TypeError("Invalid index type.")
            # check index
            if index > len(self) - 1:
                raise IndexError("Index out of range.")
            return [self.matrix[index][i] for i in range(0, index)] + [
                self.matrix[i][index] for i in range(index, len(self))
            ]
        # Handle double indexing
        elif len(item) == 2:
            row_index = None
            col_index = None
            if all(isinstance(i, int) for i in item):
                row_index, col_index = item
            elif all(isinstance(i, str) for i in item):
                row_name, col_name = item
                if row_name in self.names and col_name in self.names:
                    row_index = self.names.index(row_name)
                    col_index = self.names.index(col_name)
                else:
                    raise ValueError("Item not found.")
            else:
                raise TypeError("Invalid index type.")
            # check index
            if row_index > len(self) - 1 or col_index > len(self) - 1:
                raise IndexError("Index out of range.")
            if row_index > col_index:
                return self.matrix[row_index][col_index]
            else:
                return self.matrix[col_index][row_index]
        else:
            raise TypeError("Invalid index type.")

    def __setitem__(self, item, value):
        """Set value by the index(s) or name(s).

        Similar to __getitem__::

            dm[1] = [1, 0, 3, 4]    set values from '1' to others;
            dm[i, j] = 2            set the value from 'i' to 'j'

        """
        # Handle single indexing
        if isinstance(item, (int, str)):
            index = None
            if isinstance(item, int):
                index = item
            elif isinstance(item, str):
                if item in self.names:
                    index = self.names.index(item)
                else:
                    raise ValueError("Item not found.")
            else:
                raise TypeError("Invalid index type.")
            # check index
            if index > len(self) - 1:
                raise IndexError("Index out of range.")
            # check and assign value
            if isinstance(value, list) and all(
                isinstance(n, numbers.Number) for n in value
            ):
                if len(value) == len(self):
                    for i in range(0, index):
                        self.matrix[index][i] = value[i]
                    for i in range(index, len(self)):
                        self.matrix[i][index] = value[i]
                else:
                    raise ValueError("Value not the same size.")
            else:
                raise TypeError("Invalid value type.")
        # Handle double indexing
        elif len(item) == 2:
            row_index = None
            col_index = None
            if all(isinstance(i, int) for i in item):
                row_index, col_index = item
            elif all(isinstance(i, str) for i in item):
                row_name, col_name = item
                if row_name in self.names and col_name in self.names:
                    row_index = self.names.index(row_name)
                    col_index = self.names.index(col_name)
                else:
                    raise ValueError("Item not found.")
            else:
                raise TypeError("Invalid index type.")
            # check index
            if row_index > len(self) - 1 or col_index > len(self) - 1:
                raise IndexError("Index out of range.")
            # check and assign value
            if isinstance(value, numbers.Number):
                if row_index > col_index:
                    self.matrix[row_index][col_index] = value
                else:
                    self.matrix[col_index][row_index] = value
            else:
                raise TypeError("Invalid value type.")
        else:
            raise TypeError("Invalid index type.")

    def __delitem__(self, item):
        """Delete related distances by the index or name."""
        index = None
        if isinstance(item, int):
            index = item
        elif isinstance(item, str):
            index = self.names.index(item)
        else:
            raise TypeError("Invalid index type.")
        # remove distances related to index
        for i in range(index + 1, len(self)):
            del self.matrix[i][index]
        del self.matrix[index]
        # remove name
        del self.names[index]

    def insert(self, name, value, index=None):
        """Insert distances given the name and value.

        :Parameters:
            name : str
                name of a row/col to be inserted
            value : list
                a row/col of values to be inserted

        """
        if isinstance(name, str):
            # insert at the given index or at the end
            if index is None:
                index = len(self)
            if not isinstance(index, int):
                raise TypeError("Invalid index type.")
            # insert name
            self.names.insert(index, name)
            # insert elements of 0, to be assigned
            self.matrix.insert(index, [0] * index)
            for i in range(index, len(self)):
                self.matrix[i].insert(index, 0)
            # assign value
            self[index] = value
        else:
            raise TypeError("Invalid name type.")

    def __len__(self):
        """Matrix length."""
        return len(self.names)

    def __repr__(self):
        """Return Matrix as a string."""
        return self.__class__.__name__ + "(names=%s, matrix=%s)" % tuple(
            map(repr, (self.names, self.matrix))
        )

    def __str__(self):
        """Get a lower triangular matrix string."""
        matrix_string = "\n".join(
            [
                self.names[i] + "\t" + "\t".join([str(n) for n in self.matrix[i]])
                for i in range(0, len(self))
            ]
        )
        matrix_string = matrix_string + "\n\t" + "\t".join(self.names)
        return matrix_string

class TreeConstructor:
    """Base class for all tree constructor."""

    def build_tree(self, msa):
        """Caller to built the tree from a MultipleSeqAlignment object.

        This should be implemented in subclass.
        """
        raise NotImplementedError("Method not implemented!")


class DistanceTreeConstructor(TreeConstructor):
    """Distance based tree constructor.

    :Parameters:
        method : str
            Distance tree construction method, 'nj'(default) or 'upgma'.

    Examples
    --------
    Loading a small PHYLIP alignment from which to compute distances, and then
    build a upgma Tree::

        from Bio.Phylo.TreeConstruction import DistanceTreeConstructor
        from Bio.Phylo.TreeConstruction import DistanceCalculator
        from Bio import AlignIO
        aln = AlignIO.read(open('TreeConstruction/msa.phy'), 'phylip')
        constructor = DistanceTreeConstructor()
        calculator = DistanceCalculator('identity')
        dm = calculator.get_distance(aln)
        upgmatree = constructor.upgma(dm)
        print(upgmatree)

    Output::

        Tree(rooted=True)
            Clade(branch_length=0, name='Inner4')
                Clade(branch_length=0.18749999999999994, name='Inner1')
                    Clade(branch_length=0.07692307692307693, name='Epsilon')
                    Clade(branch_length=0.07692307692307693, name='Delta')
                Clade(branch_length=0.11057692307692304, name='Inner3')
                    Clade(branch_length=0.038461538461538464, name='Inner2')
                        Clade(branch_length=0.11538461538461536, name='Gamma')
                        Clade(branch_length=0.11538461538461536, name='Beta')
                    Clade(branch_length=0.15384615384615383, name='Alpha')

    Build a NJ Tree::

        njtree = constructor.nj(dm)
        print(njtree)

    Output::

        Tree(rooted=False)
            Clade(branch_length=0, name='Inner3')
                Clade(branch_length=0.18269230769230765, name='Alpha')
                Clade(branch_length=0.04807692307692307, name='Beta')
                Clade(branch_length=0.04807692307692307, name='Inner2')
                    Clade(branch_length=0.27884615384615385, name='Inner1')
                        Clade(branch_length=0.051282051282051266, name='Epsilon')
                        Clade(branch_length=0.10256410256410259, name='Delta')
                    Clade(branch_length=0.14423076923076922, name='Gamma')

    """

    methods = ["nj", "upgma"]

    def __init__(self, method="nj"):
        """Initialize the class."""
        if isinstance(method, str) and method in self.methods:
            self.method = method
        else:
            raise TypeError(
                "Bad method: "
                + method
                + ". Available methods: "
                + ", ".join(self.methods)
            )

    def upgma(self, distance_matrix):
        """Construct and return an UPGMA tree.

        Constructs and returns an Unweighted Pair Group Method
        with Arithmetic mean (UPGMA) tree.

        :Parameters:
            distance_matrix : DistanceMatrix
                The distance matrix for tree construction.

        """
        if not isinstance(distance_matrix, DistanceMatrix):
            raise TypeError("Must provide a DistanceMatrix object.")

        # make a copy of the distance matrix to be used
        dm = copy.deepcopy(distance_matrix)
        # init terminal clades
        clades = [BaseTree.Clade(None, name) for name in dm.names]
        # init minimum index
        min_i = 0
        min_j = 0
        inner_count = 0
        while len(dm) > 1:
            min_dist = dm[1, 0]
            # find minimum index
            for i in range(1, len(dm)):
                for j in range(0, i):
                    if min_dist >= dm[i, j]:
                        min_dist = dm[i, j]
                        min_i = i
                        min_j = j

            # create clade
            clade1 = clades[min_i]
            clade2 = clades[min_j]
            inner_count += 1
            inner_clade = BaseTree.Clade(None, "Inner" + str(inner_count))
            inner_clade.clades.append(clade1)
            inner_clade.clades.append(clade2)
            # assign branch length
            clade1.branch_length = min_dist * 1.0 / 2 - self._height_of(clade1)
            clade2.branch_length = min_dist * 1.0 / 2 - self._height_of(clade2)

            # update node list
            clades[min_j] = inner_clade
            del clades[min_i]

            # rebuild distance matrix,
            # set the distances of new node at the index of min_j
            for k in range(0, len(dm)):
                if k != min_i and k != min_j:
                    dm[min_j, k] = (dm[min_i, k] + dm[min_j, k]) * 1.0 / 2

            dm.names[min_j] = "Inner" + str(inner_count)

            del dm[min_i]
        inner_clade.branch_length = 0
        return BaseTree.Tree(inner_clade)

    def nj(self, distance_matrix):
        """Construct and return a Neighbor Joining tree.

        :Parameters:
            distance_matrix : DistanceMatrix
                The distance matrix for tree construction.

        """
        if not isinstance(distance_matrix, DistanceMatrix):
            raise TypeError("Must provide a DistanceMatrix object.")

        # make a copy of the distance matrix to be used
        dm = copy.deepcopy(distance_matrix)
        # init terminal clades
        clades = [BaseTree.Clade(None, name) for name in dm.names]
        # init node distance
        node_dist = [0] * len(dm)
        # init minimum index
        min_i = 0
        min_j = 0
        inner_count = 0
        # special cases for Minimum Alignment Matrices
        if len(dm) == 1:
            root = clades[0]

            return BaseTree.Tree(root, rooted=False)
        elif len(dm) == 2:
            # minimum distance will always be [1,0]
            min_i = 1
            min_j = 0
            clade1 = clades[min_i]
            clade2 = clades[min_j]
            clade1.branch_length = dm[min_i, min_j] / 2.0
            clade2.branch_length = dm[min_i, min_j] - clade1.branch_length
            inner_clade = BaseTree.Clade(None, "Inner")
            inner_clade.clades.append(clade1)
            inner_clade.clades.append(clade2)
            clades[0] = inner_clade
            root = clades[0]

            return BaseTree.Tree(root, rooted=False)
        while len(dm) > 2:
            # calculate nodeDist
            for i in range(0, len(dm)):
                node_dist[i] = 0
                for j in range(0, len(dm)):
                    node_dist[i] += dm[i, j]
                node_dist[i] = node_dist[i] / (len(dm) - 2)

            # find minimum distance pair
            min_dist = dm[1, 0] - node_dist[1] - node_dist[0]
            min_i = 0
            min_j = 1
            for i in range(1, len(dm)):
                for j in range(0, i):
                    temp = dm[i, j] - node_dist[i] - node_dist[j]
                    if min_dist > temp:
                        min_dist = temp
                        min_i = i
                        min_j = j
            # create clade
            clade1 = clades[min_i]
            clade2 = clades[min_j]
            inner_count += 1
            inner_clade = BaseTree.Clade(None, "Inner" + str(inner_count))
            inner_clade.clades.append(clade1)
            inner_clade.clades.append(clade2)
            # assign branch length
            clade1.branch_length = (
                dm[min_i, min_j] + node_dist[min_i] - node_dist[min_j]
            ) / 2.0
            clade2.branch_length = dm[min_i, min_j] - clade1.branch_length

            # update node list
            clades[min_j] = inner_clade
            del clades[min_i]

            # rebuild distance matrix,
            # set the distances of new node at the index of min_j
            for k in range(0, len(dm)):
                if k != min_i and k != min_j:
                    dm[min_j, k] = (
                        dm[min_i, k] + dm[min_j, k] - dm[min_i, min_j]
                    ) / 2.0

            dm.names[min_j] = "Inner" + str(inner_count)
            del dm[min_i]

        # set the last clade as one of the child of the inner_clade
        root = None
        if clades[0] == inner_clade:
            clades[0].branch_length = 0
            clades[1].branch_length = dm[1, 0]
            clades[0].clades.append(clades[1])
            root = clades[0]
        else:
            clades[0].branch_length = dm[1, 0]
            clades[1].branch_length = 0
            clades[1].clades.append(clades[0])
            root = clades[1]

        return BaseTree.Tree(root, rooted=False)

    def _height_of(self, clade):
        """Calculate clade height -- the longest path to any terminal (PRIVATE)."""
        if clade.is_terminal():
            height = 0 
        else:
            height = max(self._height_of(c) + c.branch_length for c in clade.clades)

        return height




import numpy
from numpy.testing import *

from prody.tests.datafiles import *

from prody.tests import unittest, TEMPDIR

ATOMS = parseDatafile('1ubi')
ATOMS.setData('ones', numpy.zeros(len(ATOMS)))

class TestUserData(unittest.TestCase):

    def testSetData(self):

        ATOMS[:10].setData('ones', 1)
        ATOMS[10:].setData('ones', 1)
        assert_equal(ATOMS.getData('ones'), numpy.ones(len(ATOMS)))



"""This module contains unit tests for fragmenting function and methods."""

from prody.tests import TestCase

from numpy import arange
from numpy.random import shuffle

from prody import *
from prody.tests.datafiles import pathDatafile

AG = prody.parsePDB(pathDatafile('pdb3mht.pdb'))
SHUFFLED = arange(len(AG))
shuffle(SHUFFLED)
SHUFFLED = AtomMap(AG, SHUFFLED)

RTER = prody.parsePDB(pathDatafile('pdbRTER.pdb'))


class TestShuffled(TestCase):

    def testCA(self):

        self.assertEqual(AG.ca.numAtoms(), SHUFFLED.ca.numAtoms())

    def testProtein(self):

        self.assertEqual(AG.protein.numAtoms(), SHUFFLED.protein.numAtoms())


class TestTerRecord(TestCase):

    def testNumResidues(self):

        self.assertEqual(RTER.getHierView().numResidues(), 9)

    def testResidueIndexing(self):

        self.assertEqual(len(RTER.getHierView()['A', 864]), 2)

    def testSelectionResidueIndexing(self):

        residues = RTER[:32].getHierView()['A', 864]
        self.assertEqual(len(residues), 2)
        self.assertEqual((residues[0] + residues[1]).numAtoms(), 5)

    def testSelectionResidueIndexing2(self):

        self.assertEqual(len(RTER[20:].getHierView()['A', 866]), 3)


# -*- coding: utf-8 -*-
"""This module defines functions for computing structural/dynamical
domain decompositions, and related properties, from either ANM modes 
or analysis of structural ensembles.
"""

import prody
LOGGER = prody.LOGGER
SETTINGS = prody.SETTINGS

__all__ = []

from . import spectrus 
__all__.extend(spectrus.__all__)
from .spectrus import *
